import logging
import math
from ctypes import c_uint32, c_uint64, sizeof
from distutils.version import LooseVersion
from pathlib import Path
from typing import TYPE_CHECKING, Any, Dict, List, Optional, Set, Tuple, Type, TypeVar

from _ctypes import Structure

from strongarm.macho.arch_independent_structs import (
    ArchIndependentStructure,
    CFStringStruct,
    DylibCommandStruct,
    MachoBuildToolVersionStruct,
    MachoBuildVersionCommandStruct,
    MachoDyldInfoCommandStruct,
    MachoDysymtabCommandStruct,
    MachoEncryptionInfoStruct,
    MachoHeaderStruct,
    MachoLinkeditDataCommandStruct,
    MachoLoadCommandStruct,
    MachoNlistStruct,
    MachoSectionRawStruct,
    MachoSegmentCommandStruct,
    MachoSymtabCommandStruct,
)
from strongarm.macho.macho_definitions import (
    CPU_TYPE,
    HEADER_FLAGS,
    DylibCommand,
    DylibStruct,
    LcStrUnion,
    MachArch,
    MachoBuildVersionPlatform,
    MachoFileType,
    StaticFilePointer,
    VirtualMemoryPointer,
)
from strongarm.macho.macho_load_commands import MachoLoadCommands

if TYPE_CHECKING:
    from strongarm.macho.codesign import CodesignParser

AIS = TypeVar("AIS", bound=ArchIndependentStructure)


class BinaryEncryptedError(Exception):
    """Raised when the binary is encrypted.
    """


class LoadCommandMissingError(Exception):
    """Raised when the binary is missing a load command.
    """


class NoEmptySpaceForLoadCommandError(Exception):
    """Raised when we fail to insert a load command because there's not enough empty space left in the Mach-O header.
    """


class InvalidAddressError(Exception):
    """Raised when a client asks for bytes at an address outside the binary.
    """


class MachoSegment:
    def __init__(self, binary: "MachoBinary", segment_command: MachoSegmentCommandStruct) -> None:
        self.cmd = segment_command

        self.name = segment_command.segname.decode()
        self.sizeof = segment_command.sizeof

        self.vmaddr = segment_command.vmaddr
        self.vmsize = segment_command.vmsize
        self.vm_end_address = self.vmaddr + self.vmsize

        self.offset = segment_command.fileoff
        self.size = segment_command.filesize
        self.end_address = self.offset + self.size

        self.section_count = segment_command.nsects
        self.sections: List["MachoSection"] = []

        self.maxprot = segment_command.maxprot
        self.initprot = segment_command.initprot
        self.flags = segment_command.flags

    def __repr__(self) -> str:
        virtual_loc = f"[0x{self.vmaddr:011x} - 0x{self.vm_end_address:011x}]"
        file_loc = f"[0x{self.offset:011x} - 0x{self.end_address:011x}]"
        return f"<MachoSegment {virtual_loc} (file {file_loc}) {self.name} ({self.section_count} sections)>"


class MachoSection:
    def __init__(self, binary: "MachoBinary", section_command: MachoSectionRawStruct, segment: MachoSegment) -> None:
        self.cmd = section_command
        self.segment = segment

        # ignore these types due to dynamic attributes of associated types
        self.name = section_command.sectname.decode()
        self.segment_name = section_command.segname.decode()
        self.address = section_command.addr
        self.size = section_command.size
        self.end_address = self.address + self.size
        self.offset = section_command.offset

        self.align = section_command.align
        self.reloff = section_command.reloff
        self.nreloc = section_command.nreloc
        self.flags = section_command.flags

    def __repr__(self) -> str:
        virtual_loc = f"[0x{self.address:011x} - 0x{self.end_address:011x}]"
        return f'<MachoSection {virtual_loc} "{self.name}" ("{self.segment_name}")>'


class MachoBinary:
    _MAG_64 = [MachArch.MH_MAGIC_64, MachArch.MH_CIGAM_64]
    _MAG_32 = [MachArch.MH_MAGIC, MachArch.MH_CIGAM]
    _MAG_BIG_ENDIAN = [MachArch.MH_CIGAM, MachArch.MH_CIGAM_64]
    SUPPORTED_MAG = _MAG_64 + _MAG_32
    BYTES_PER_INSTRUCTION = 4

    def __init__(self, path: Path, binary_data: bytes, file_offset: Optional[StaticFilePointer] = None,) -> None:
        """Parse the bytes representing a Mach-O file.
        """
        from .codesign.codesign_parser import CodesignParser

        self._cached_binary = binary_data

        self.path = path
        self.is_64bit: bool = False
        self.is_swap: bool = False
        self.slice_filesize = len(binary_data)
        self._load_commands_end_addr = 0
        self.file_offset = file_offset or StaticFilePointer(0x0)

        # Mach-O header data
        self.cpu_type: CPU_TYPE = CPU_TYPE.UNKNOWN  # Overwritten later in the parse
        self._header: Optional[MachoHeaderStruct] = None
        self.header_flags: List[int] = []
        self.file_type: MachoFileType = MachoFileType.MH_EXECUTE  # Overwritten later in the parse
        self._virtual_base: Optional[VirtualMemoryPointer] = None

        # Segment and section commands from Mach-O header
        self.segments: List[MachoSegment] = []
        self.sections: List[MachoSection] = []
        # Interesting Mach-O sections
        self._dysymtab: Optional[MachoDysymtabCommandStruct] = None
        self._symtab: Optional[MachoSymtabCommandStruct] = None
        self._encryption_info: Optional[MachoEncryptionInfoStruct] = None
        self._dyld_info: Optional[MachoDyldInfoCommandStruct] = None
        self._dyld_export_trie: Optional[MachoLinkeditDataCommandStruct] = None
        self._dyld_chained_fixups: Optional[MachoLinkeditDataCommandStruct] = None
        self.load_dylib_commands: List[DylibCommandStruct] = []
        self._code_signature_cmd: Optional[MachoLinkeditDataCommandStruct] = None
        self._function_starts_cmd: Optional[MachoLinkeditDataCommandStruct] = None
        self._functions_list: Optional[Set[VirtualMemoryPointer]] = None
        self._id_dylib_cmd: Optional[DylibCommandStruct] = None
        self._build_version_cmd: Optional[MachoBuildVersionCommandStruct] = None
        self._build_tool_versions: Optional[List[MachoBuildToolVersionStruct]] = None

        self.__codesign_parser: Optional[CodesignParser] = None
        self.__minimum_deployment_target: Optional[LooseVersion] = None

        # This kicks off the parse of the binary
        if not self.parse():
            raise RuntimeError("Failed to parse Mach-O")

        self.platform_word_type = c_uint64 if self.is_64bit else c_uint32

        self._symtab_contents: Optional[List[MachoNlistStruct]] = None

        from .dyld_info_parser import DyldBoundSymbol, DyldInfoParser

        self.dyld_bound_symbols: Dict[VirtualMemoryPointer, DyldBoundSymbol] = {}
        self.dyld_rebased_pointers: Dict[VirtualMemoryPointer, VirtualMemoryPointer] = {}

        if self._dyld_chained_fixups:
            rebases, binds = DyldInfoParser.parse_chained_fixups(self)  # type: ignore
            self.dyld_rebased_pointers, self.dyld_bound_symbols = rebases, binds
        else:
            self.dyld_bound_symbols = DyldInfoParser.parse_dyld_info(self)

    def __repr__(self) -> str:
        return f"<MachoBinary binary={self.path}>"

    def parse(self) -> bool:
        """Attempt to parse the provided file info as a Mach-O slice

        Returns:
            True if the file data represents a valid & supported Mach-O which was successfully parsed.
            False otherwise.

        """
        # logging.debug(self, f"parsing Mach-O slice @ {hex(int(self._offset_within_fat))} in {self.path}")

        # Preliminary Mach-O parsing
        if not self.verify_magic():
            logging.debug(self, f"unsupported magic {hex(self.slice_magic)}")
            return False

        self.is_swap = self.should_swap_bytes()
        # Big endian binaries are currently unsupported
        if self.is_swap:
            raise NotImplementedError("Big-endian binaries are unsupported")

        self.is_64bit = self.magic_is_64()

        self.parse_header()

        logging.debug(self, f"header parsed. non-native endianness? {self.is_swap}. 64-bit? {self.is_64bit}")
        return True

    @property
    def slice_magic(self) -> int:
        """Read magic number identifier from this Mach-O slice
        """
        return self.read_word(0, virtual=False, word_type=c_uint32)

    def verify_magic(self) -> bool:
        """Ensure magic at beginning of Mach-O slice indicates a supported format

        Returns:
            True if the magic represents a supported file format, False if the magic represents an unsupported format

        """
        return self.slice_magic in MachoBinary.SUPPORTED_MAG

    def magic_is_64(self) -> bool:
        """Convenience method to check if our magic corresponds to a 64-bit slice

        Returns:
            True if self.slice_magic corresponds to a 64 bit MachO slice, False otherwise

        """
        return self.slice_magic in MachoBinary._MAG_64

    def parse_header(self) -> None:
        """Read all relevant info from a Mach-O header which does not require cross-referencing.
        Specifically, this method parses the Mach-O header & header flags, CPU target,
        and all segment and section commands.
        """
        self._header = self.read_struct(0, MachoHeaderStruct)

        if self.header.cputype == MachArch.MH_CPU_TYPE_ARM:  # type: ignore
            self.cpu_type = CPU_TYPE.ARMV7
        elif self.header.cputype == MachArch.MH_CPU_TYPE_ARM64:  # type: ignore
            self.cpu_type = CPU_TYPE.ARM64
        else:
            self.cpu_type = CPU_TYPE.UNKNOWN

        self._parse_header_flags()
        self.file_type = MachoFileType(self.header.filetype)

        # load commands begin directly after Mach O header, so the offset is the size of the header
        load_commands_off = self.header.sizeof

        self._load_commands_end_addr = load_commands_off + self.header.sizeofcmds  # type: ignore
        self._parse_load_commands(load_commands_off, self.header.ncmds)  # type: ignore

    def _parse_header_flags(self) -> None:
        """Interpret binary's header bitset and populate self.header_flags
        """
        self.header_flags = []

        flags_bitset = self.header.flags
        for mask in [x.value for x in HEADER_FLAGS]:
            # is this mask set in the binary's flags?
            if (flags_bitset & mask) == mask:
                # mask is present in bitset, add to list of included flags
                self.header_flags.append(mask)

    def _parse_load_commands(self, offset: StaticFilePointer, ncmds: int) -> None:
        """Parse Mach-O segment commands beginning at a given slice offset

        Args:
            offset: Slice offset to first segment command
            ncmds: Number of load commands to parse, as declared by the header's ncmds field
        """
        self.load_dylib_commands = []

        for i in range(ncmds):
            load_command = self.read_struct(offset, MachoLoadCommandStruct)

            if load_command.cmd in [MachoLoadCommands.LC_SEGMENT, MachoLoadCommands.LC_SEGMENT_64]:
                segment_command = self.read_struct(offset, MachoSegmentCommandStruct)
                # TODO(PT) handle byte swap of segment if necessary
                segment = MachoSegment(self, segment_command)
                self.segments.append(segment)
                self._parse_sections_for_segment(segment, offset)

            # some commands have their own structure that we interpret separately from a normal load command
            # if we want to interpret more commands in the future, this is the place to do it
            elif load_command.cmd in [MachoLoadCommands.LC_ENCRYPTION_INFO, MachoLoadCommands.LC_ENCRYPTION_INFO_64]:
                self._encryption_info = self.read_struct(offset, MachoEncryptionInfoStruct)

            elif load_command.cmd == MachoLoadCommands.LC_SYMTAB:
                self._symtab = self.read_struct(offset, MachoSymtabCommandStruct)

            elif load_command.cmd == MachoLoadCommands.LC_DYSYMTAB:
                self._dysymtab = self.read_struct(offset, MachoDysymtabCommandStruct)

            elif load_command.cmd in [MachoLoadCommands.LC_DYLD_INFO, MachoLoadCommands.LC_DYLD_INFO_ONLY]:
                self._dyld_info = self.read_struct(offset, MachoDyldInfoCommandStruct)

            elif load_command.cmd in [MachoLoadCommands.LC_DYLD_EXPORTS_TRIE]:
                self._dyld_export_trie = self.read_struct(offset, MachoLinkeditDataCommandStruct)

            elif load_command.cmd in [MachoLoadCommands.LC_DYLD_CHAINED_FIXUPS]:
                self._dyld_chained_fixups = self.read_struct(offset, MachoLinkeditDataCommandStruct)

            elif load_command.cmd in [MachoLoadCommands.LC_LOAD_DYLIB, MachoLoadCommands.LC_LOAD_WEAK_DYLIB]:
                dylib_load_command = self.read_struct(offset, DylibCommandStruct)
                self.load_dylib_commands.append(dylib_load_command)

            elif load_command.cmd == MachoLoadCommands.LC_CODE_SIGNATURE:
                self._code_signature_cmd = self.read_struct(offset, MachoLinkeditDataCommandStruct)

            elif load_command.cmd == MachoLoadCommands.LC_FUNCTION_STARTS:
                self._function_starts_cmd = self.read_struct(offset, MachoLinkeditDataCommandStruct)

            elif load_command.cmd == MachoLoadCommands.LC_ID_DYLIB:
                self._id_dylib_cmd = self.read_struct(offset, DylibCommandStruct)
                # This load command should only be present for dylibs. Validate this assumption
                assert self.file_type == MachoFileType.MH_DYLIB

            elif load_command.cmd == MachoLoadCommands.LC_BUILD_VERSION:
                self._build_version_cmd = self.read_struct(offset, MachoBuildVersionCommandStruct)
                # Parse the build tool versions following this structure
                build_tool_offset = offset + self._build_version_cmd.sizeof
                self._build_tool_versions = []
                for _ in range(self._build_version_cmd.ntools):
                    build_tool_version = self.read_struct(build_tool_offset, MachoBuildToolVersionStruct)
                    self._build_tool_versions.append(build_tool_version)

            # move to next load command in header
            offset += load_command.cmdsize

    def read_struct(self, binary_offset: int, struct_type: Type[AIS], virtual: bool = False) -> AIS:
        """Given an binary offset, return the structure it describes.

        Params:
            binary_offset: Address from where to read the bytes.
            struct_type: ArchIndependentStructure subclass.
            virtual: Whether the address should be slid (virtual) or not.
        Returns:
            ArchIndependentStructure loaded from the pointed address.
        """
        backing_layout = struct_type.get_backing_data_layout(self.is_64bit, self.get_minimum_deployment_target())
        data = self.get_contents_from_address(address=binary_offset, size=sizeof(backing_layout), is_virtual=virtual)
        return struct_type(binary_offset, data, backing_layout)

    def read_struct_with_rebased_pointers(
        self, binary_offset: int, struct_type: Type[AIS], virtual: bool = False
    ) -> AIS:
        """Read a static binary structure that may contain rebased pointers.
        For each uint64_t within the structure, check if we know that this pointer should be rebased.
        If so, the static data here may be a packed chained fixup pointer, rather than a pointer we can follow.
        In this case, update the pointer to contain the value to be rebased, so that the pointer can be followed.
        """
        backing_layout = struct_type.get_backing_data_layout(self.is_64bit, self.get_minimum_deployment_target())
        data = self.get_contents_from_address(address=binary_offset, size=sizeof(backing_layout), is_virtual=virtual)
        s = struct_type(binary_offset, data, backing_layout)

        # Check each c_uint64/c_ulong to see if we know about a rebase for it. If so, apply it
        base_virt_offset = binary_offset
        if not virtual:
            base_virt_offset += self.get_virtual_base()
        for field_name, field_type, *_ in backing_layout._fields_:
            field_offset = getattr(getattr(backing_layout, field_name), "offset")
            field_address = base_virt_offset + field_offset
            if field_type == c_uint64 and field_address in self.dyld_rebased_pointers:
                logging.debug(
                    f"Setting rebased pointer within {struct_type}+{field_offset} -> "
                    f"{self.dyld_rebased_pointers[field_address]} at {field_address}"
                )
                setattr(s, field_name, self.dyld_rebased_pointers[field_address])

        return s

    def section_name_for_address(self, virt_addr: VirtualMemoryPointer) -> Optional[str]:
        """Given an address in the virtual address space, return the name of the section which contains it.
        """
        section = self.section_for_address(virt_addr)
        if not section:
            return None
        return section.name

    def section_for_address(self, virt_addr: VirtualMemoryPointer) -> Optional[MachoSection]:
        """Given an address in the virtual address space, return the section which contains it.
        """
        # invalid address?
        if virt_addr < self.get_virtual_base():
            return None

        # if the address given is past the last declared section, translate based on the last section
        # so, we need to keep track of the last seen section
        max_section = next(iter(self.sections))

        for idx, section in enumerate(self.sections):
            # update highest section
            if section.address > max_section.address:
                max_section = section

            if section.address <= virt_addr < section.end_address:
                return section
        # we looked through all sections and didn't find one explicitly containing this address
        # guess by using the highest-addressed section we've seen
        return max_section

    def segment_for_index(self, segment_index: int) -> MachoSegment:
        if 0 <= segment_index < len(self.segments):
            # PT: Segments are guaranteed to be sorted in the order they appear in the Mach-O header
            return self.segments[segment_index]
        else:
            raise ValueError(f"segment_index ({segment_index}) out of bounds ({len(self.segments)}")

    def segment_with_name(self, desired_segment_name: str) -> Optional[MachoSegment]:
        """Returns the segment with the provided name. Returns None if there's no such segment in the binary.
        """
        # TODO(PT): add unit test for this method
        return next((s for s in self.segments if s.name == desired_segment_name), None)

    def section_with_name(self, desired_section_name: str, parent_segment_name: str) -> Optional[MachoSection]:
        """Retrieve the section with the provided name which is contained within the provided segment.
        Returns None if no such section exists.
        """
        segment = self.segment_with_name(parent_segment_name)
        if segment:
            return next((s for s in segment.sections if s.name == desired_section_name), None)
        return None

    def _parse_sections_for_segment(self, segment: MachoSegment, segment_offset: StaticFilePointer) -> None:
        """Parse all sections contained within a Mach-O segment, and add them to our list of sections

        Args:
            segment: The segment command whose sections should be read
            segment_offset: The offset within the file that the segment command is located at
        """
        if not segment.section_count:
            return

        # The first section of this segment begins directly after the segment
        section_offset = segment_offset + segment.sizeof
        for i in range(segment.section_count):
            # Read section header from file
            # TODO(PT): handle byte swap of segment
            section_command = self.read_struct(section_offset, MachoSectionRawStruct)
            # Encapsulate header and content into one object, and store that
            section = MachoSection(self, section_command, segment)
            segment.sections.append(section)
            # Add to list of sections within the Mach-O
            self.sections.append(section)

            # Iterate to next section in list
            section_offset += section_command.sizeof

    def get_virtual_base(self) -> VirtualMemoryPointer:
        """Retrieve the first virtual address of the Mach-O slice

        Returns:
            int containing the virtual memory space address that the Mach-O slice requests to begin at

        """
        if not self._virtual_base:
            text_seg = self.segment_with_name("__TEXT")
            if not text_seg:
                raise RuntimeError("Could not find virtual base because binary has no __TEXT segment.")
            self._virtual_base = VirtualMemoryPointer(text_seg.vmaddr)

        return self._virtual_base

    def get_file_offset(self) -> StaticFilePointer:
        """Retrieve the offset within the file of this Mach-O slice
        """
        return self.file_offset

    def get_bytes(self, offset: StaticFilePointer, size: int, _translate_addr_to_file: bool = False) -> bytearray:
        """Retrieve bytes from Mach-O slice, taking into account that the slice could be at an offset within a FAT

        Args:
            offset: index from beginning of slice to retrieve data from
            size: maximum number of bytes to read
            _translate_addr_to_file: Internal option to support parsing DYLD shared cache binaries.
                Images within the shared cache store some of their data in a separate part of the cache from their code.
                This option tells DYLD cache binaries to translate the offset into the global cache file.

        Returns:
            string containing byte content of mach-o slice at an offset from the start of the slice

        """
        if offset > 0x100000000:
            raise InvalidAddressError(
                f"get_bytes() offset {hex(offset)} looks like a virtual address."
                " Did you mean to use get_content_from_virtual_address?"
            )
        if offset < 0:
            raise InvalidAddressError(f"get_bytes() passed negative offset: {hex(offset)}")
        if _translate_addr_to_file:
            raise ValueError("_translate_addr_to_file may only be used with dyld_shared_cache binaries")

        # safeguard against reading from an encrypted segment of the binary
        if self.is_range_encrypted(offset, size):
            encryption_range_start = int(self.encryption_info.cryptoff)
            encryption_range_end = encryption_range_start + int(self.encryption_info.cryptsize)
            raise BinaryEncryptedError(
                f"Cannot read encrypted range [{hex(encryption_range_start)} - {hex(encryption_range_end)}]"
            )

        return bytearray(self._cached_binary[offset : offset + size])

    def should_swap_bytes(self) -> bool:
        """Check whether self.slice_magic refers to a big-endian Mach-O binary

        Returns:
            True if self.slice_magic indicates a big endian Mach-O, False otherwise

        """
        return self.slice_magic in MachoBinary._MAG_BIG_ENDIAN

    def get_raw_string_table(self) -> List[int]:
        """Read string table from binary, as described by LC_SYMTAB. Each strtab entry is terminated
        by a NULL character.

        Returns:
            Raw, packed array of characters containing binary's string table data

        """
        string_table_data = self.get_bytes(self.symtab.stroff, self.symtab.strsize)
        # split into characters (string table is packed and each entry is terminated by a null character)
        string_table = list(string_table_data)
        return string_table

    @property
    def symtab_contents(self) -> List[MachoNlistStruct]:
        if self._symtab_contents is None:
            self._symtab_contents = self._parse_symtab_contents()
            logging.debug(self, f"parsed symtab, len = {len(self.symtab_contents)}")
        return self._symtab_contents

    def _parse_symtab_contents(self) -> List[MachoNlistStruct]:
        """Parse symbol table containing list of Nlist64's

        Returns:
            Array of Nlist64's representing binary's symbol table

        """
        logging.debug(self, f"parsing {self.symtab.nsyms} symtab entries")

        symtab = []
        # start reading from symoff and increment by one Nlist64 each iteration
        symoff = self.symtab.symoff
        for i in range(self.symtab.nsyms):
            nlist = self.read_struct(symoff, MachoNlistStruct)
            symtab.append(nlist)
            # go to next Nlist in file
            symoff += nlist.sizeof

        return symtab

    def get_indirect_symbol_table(self) -> List[int]:
        indirect_symtab = []
        # dysymtab has fields that tell us the file offset of the indirect symbol table, as well as the number
        # of indirect symbols present in the mach-o
        indirect_symtab_off = self.dysymtab.indirectsymoff

        # indirect symtab is an array of uint32's
        for i in range(self.dysymtab.nindirectsyms):
            indirect_symtab_entry = self.read_word(indirect_symtab_off, virtual=False, word_type=c_uint32)
            indirect_symtab.append(int(indirect_symtab_entry))
            # traverse to next pointer
            indirect_symtab_off += sizeof(c_uint32)
        return indirect_symtab

    def file_offset_for_virtual_address(self, virtual_address: VirtualMemoryPointer) -> StaticFilePointer:
        # if this address is within the initial Mach-O load commands, it must be handled seperately
        # this unslid virtual address is just a 'best guess' of the physical file address, and it'll be the correct
        # address if the virtual address was within the initial load commands
        # if the virtual address was in the section contents, however, we must use another method to translate addresses
        unslid_virtual_address = virtual_address - self.get_virtual_base()
        if unslid_virtual_address < self._load_commands_end_addr:
            return StaticFilePointer(unslid_virtual_address)

        section_for_address = self.section_for_address(virtual_address)
        if not section_for_address:
            raise RuntimeError(f"Could not map virtual address {hex(int(virtual_address))} to a section!")

        # the virtual address is contained within a section's contents
        # use this formula to convert a virtual address within a section to the file offset:
        # https://reverseengineering.stackexchange.com/questions/8177/convert-mach-o-vm-address-to-file-offset
        binary_address = (virtual_address - section_for_address.address) + section_for_address.offset
        return StaticFilePointer(binary_address)

    def get_content_from_virtual_address(self, virtual_address: VirtualMemoryPointer, size: int) -> bytearray:
        binary_address = self.file_offset_for_virtual_address(virtual_address)
        return self.get_bytes(binary_address, size)

    def get_contents_from_address(self, address: int, size: int, is_virtual: bool = False) -> bytearray:
        """Get a bytesarray from a specified address, size and virtualness
        TODO(FS): change all methods that use addresses as ints to the VirtualAddress/StaticAddress class pair to better
         express intent
        """
        if is_virtual:
            return self.get_content_from_virtual_address(VirtualMemoryPointer(address), size)
        else:
            return self.get_bytes(StaticFilePointer(address), size)

    def get_full_string_from_start_address(self, start_address: int, virtual: bool = True) -> Optional[str]:
        """Return a string containing the bytes from start_address up to the next NULL character
        This method will return None if the specified address does not point to a UTF-8 encoded string
        """
        max_len = 16
        symbol_name_characters = []
        found_null_terminator = False

        while not found_null_terminator:
            if virtual:
                name_bytes = self.get_content_from_virtual_address(VirtualMemoryPointer(start_address), max_len)
            else:
                name_bytes = self.get_bytes(StaticFilePointer(start_address), max_len)
            # search for null terminator in this content
            for ch in name_bytes:
                if ch == 0x00:
                    found_null_terminator = True
                    break
                symbol_name_characters.append(ch)

            # do we need to keep searching for the end of the symbol name?
            if not found_null_terminator:
                # since we read [start_address:start_address + max_len], trim that from search space
                start_address += max_len
                # double search space for next iteration
                max_len *= 2
            else:
                # read full string!
                try:
                    symbol_name = bytearray(symbol_name_characters).decode()
                    return symbol_name
                except UnicodeDecodeError:
                    # if decoding the string failed, we may have been passed an address which does not actually
                    # point to a string
                    return None
        return None

    def read_string_at_address(self, address: VirtualMemoryPointer) -> Optional[str]:
        """Read a string embedded in the binary at address
        This method will automatically parse a CFString and return the string literal if address points to one
        """
        section_name = self.section_name_for_address(address)
        # no section found?
        if not section_name:
            return None
        # some sections will never contain a string literal and can cause errors if we try to read a string from them
        if section_name in ["__bss", "__objc_selrefs", "__objc_classrefs"]:
            return None
        # special case if this is a __cfstring entry
        if section_name == "__cfstring":
            # read bytes into CFString struct
            cfstring_ent = self.read_struct(address, CFStringStruct, virtual=True)
            # patch address to read string from to be the string literal address of this CFString
            address = cfstring_ent.literal
        return self.get_full_string_from_start_address(address)

    def is_encrypted(self) -> bool:
        """Returns True if the binary has an encrypted segment, False otherwise
        """
        if not self._encryption_info:
            return False
        return self.encryption_info.cryptid != 0

    def is_range_encrypted(self, offset: StaticFilePointer, size: int) -> bool:
        """Returns whether the provided address range overlaps with the encrypted section of the binary.
        """
        if not self.is_encrypted():
            return False

        # if 2 ranges overlap, the end address of the first range will be greater than the start of the second, and
        # the end address of the second will be greater than the start of the first
        range1 = (offset, offset + size)
        range2 = (self.encryption_info.cryptoff, self.encryption_info.cryptoff + self.encryption_info.cryptsize)
        return range1[1] >= range2[0] and range2[1] >= range1[0]

    def dylib_for_library_ordinal(self, library_ordinal: int) -> Optional[DylibCommandStruct]:
        """Retrieve the library information for the 'library ordinal' value, or None if no entry exists there.
        Library ordinals are 1-indexed.

        https://opensource.apple.com/source/cctools/cctools-795/include/mach-o/loader.h
        """
        idx = library_ordinal - 1
        # library ordinals are 1-indexed
        # if the input is invalid, return None
        if library_ordinal < 1 or idx >= len(self.load_dylib_commands):
            return None
        return self.load_dylib_commands[idx]

    def dylib_name_for_library_ordinal(self, library_ordinal: int) -> str:
        """Read the name of the dynamic library by its library ordinal
        """
        source_dylib = self.dylib_for_library_ordinal(library_ordinal)
        if source_dylib:
            source_name_addr = source_dylib.binary_offset + source_dylib.dylib.name.offset + self.get_virtual_base()
            source_name = self.get_full_string_from_start_address(source_name_addr)
            if not source_name:
                source_name = "<unknown dylib>"
        else:
            # we have encountered binaries where the n_desc indicates a nonexistent library ordinal
            # Netflix.app/frameworks/widevine_cdm_sdk_oemcrypto_release.framework/widevine_cdm_sdk_oemcrypto_release
            # indicates an ordinal 254, when the binary only actually has 8 LC_LOAD_DYLIB commands.
            # if we encounter a buggy binary like this, just use a placeholder name
            source_name = "<unknown dylib>"
        return source_name

    def read_pointer_section(self, section_name: str) -> Tuple[List[VirtualMemoryPointer], List[VirtualMemoryPointer]]:
        """Read all the pointers in a section

        It is the caller's responsibility to only call this with a `section_name` which indicates a section which should
        only contain a pointer list.

        The return value is two lists of pointers.
        The first List contains the virtual addresses of each entry in the section.
        The second List contains the pointer values contained at each of these addresses.

        The indexes of these two lists are matched up; that is, list1[0] is the virtual address of the first pointer
        in the requested section, and list2[0] is the pointer value contained at that address.
        """
        locations: List[VirtualMemoryPointer] = []
        entries: List[VirtualMemoryPointer] = []

        # PT: Assume a pointer-list-section will always be in __DATA or __DATA_CONST. True as far as I know.
        for segment in ["__DATA", "__DATA_CONST"]:
            section = self.section_with_name(section_name, segment)
            if section:
                break
        else:
            # Couldn't find the desired section
            return locations, entries

        section_base = section.address
        section_data = self.get_bytes(section.offset, section.size)

        binary_word = self.platform_word_type
        pointer_count = int(len(section_data) / sizeof(binary_word))
        pointer_off = 0

        for i in range(pointer_count):
            # convert section offset of entry to absolute virtual address
            ptr_location = VirtualMemoryPointer(section_base + pointer_off)
            locations.append(ptr_location)

            if ptr_location in self.dyld_rebased_pointers:
                ptr_value = self.dyld_rebased_pointers[ptr_location]
                logging.debug(f"Pointer is rebased: {ptr_location} -> {ptr_value}")
            else:
                data_end = pointer_off + sizeof(binary_word)
                ptr_value = VirtualMemoryPointer(
                    binary_word.from_buffer(bytearray(section_data[pointer_off:data_end])).value
                )
                logging.debug(f"Pointer was not in the rebase list: {ptr_location} -> {ptr_value}")

            entries.append(VirtualMemoryPointer(ptr_value))

            pointer_off += sizeof(binary_word)

        return locations, entries

    def read_word(self, address: int, virtual: bool = True, word_type: Any = None) -> int:
        """Attempt to read a word from the binary at a virtual address.
        """
        if not word_type:
            word_type = self.platform_word_type

        if virtual:
            file_bytes = self.get_content_from_virtual_address(VirtualMemoryPointer(address), sizeof(word_type))
        else:
            file_bytes = self.get_bytes(StaticFilePointer(address), sizeof(word_type))

        if not file_bytes:
            raise InvalidAddressError(f"Could not read word at address {hex(address)}")

        return word_type.from_buffer(bytearray(file_bytes)).value

    def read_rebased_pointer(self, address: VirtualMemoryPointer) -> VirtualMemoryPointer:
        """Attempt to read a rebased pointer from the binary at a virtual address.
        The pointer is assumed to be the platform word size.
        """
        if address not in self.dyld_rebased_pointers:
            # This may be a pre-iOS 15 binary for which we don't record rebases
            return VirtualMemoryPointer(self.read_word(address, virtual=True, word_type=self.platform_word_type))

        return self.dyld_rebased_pointers[address]

    @property
    def header(self) -> MachoHeaderStruct:
        if self._header:
            return self._header
        else:
            raise LoadCommandMissingError()

    @property
    def dysymtab(self) -> MachoDysymtabCommandStruct:
        if self._dysymtab:
            return self._dysymtab
        else:
            raise LoadCommandMissingError()

    @property
    def symtab(self) -> MachoSymtabCommandStruct:
        if self._symtab:
            return self._symtab
        else:
            raise LoadCommandMissingError()

    @property
    def encryption_info(self) -> MachoEncryptionInfoStruct:
        if self._encryption_info:
            return self._encryption_info
        else:
            raise LoadCommandMissingError()

    @property
    def dyld_info(self) -> MachoDyldInfoCommandStruct:
        if self._dyld_info:
            return self._dyld_info
        else:
            raise LoadCommandMissingError()

    @property
    def code_signature_cmd(self) -> Optional[MachoLinkeditDataCommandStruct]:
        return self._code_signature_cmd

    @property
    def _codesign_parser(self) -> "CodesignParser":
        if not self.__codesign_parser:
            from strongarm.macho.codesign import CodesignParser

            self.__codesign_parser = CodesignParser(self)
        return self.__codesign_parser

    def get_entitlements(self) -> Optional[bytearray]:
        """Read the entitlements the binary was signed with.
        """
        return self._codesign_parser.entitlements

    def get_signing_identity(self) -> Optional[str]:
        """Read the bundle ID the binary was signed as.
        """
        return self._codesign_parser.signing_identifier

    def get_team_id(self) -> Optional[str]:
        """Read the team ID the binary was signed with.
        """
        return self._codesign_parser.signing_team_id

    def write_bytes(self, data: bytes, address: int, virtual: bool = False) -> "MachoBinary":
        """Overwrite the data in the current binary with the provided data, returning a new modified binary.
        Note: This will invalidate the binary's code signature, if present.
        """
        # Ensure there is valid data in this address region by trying to read from it
        self.get_contents_from_address(address, len(data), virtual)

        # If the above did not throw an exception, the provided address range is valid.
        file_offset = address
        if virtual:
            file_offset = self.file_offset_for_virtual_address(VirtualMemoryPointer(address))

        # Create a new binary with the overwritten data
        new_binary_data = bytearray(len(self._cached_binary))
        new_binary_data[:] = self._cached_binary
        new_binary_data[file_offset : file_offset + len(data)] = data

        return MachoBinary(self.path, new_binary_data)

    def write_struct(self, struct: Structure, address: int, virtual: bool = False) -> "MachoBinary":
        """Serialize and write the provided structure the Mach-O slice, returning a new modified binary.
        Note: This will invalidate the binary's code signature, if present.
        TODO(PT): Deprecate and move to MachoBinaryWriter
        """
        # Write the structure bytes to the binary
        # TODO(PT): byte order?
        return self.write_bytes(bytes(struct), address, virtual)

    def insert_load_dylib_cmd(self, dylib_path: str) -> "MachoBinary":
        """Add a load command of the provided dylib path to the Mach-O header, returning a modified binary.
        This will increase mh_header->ncmds by 1, and mh_header->sizeofcmds by the size of the new load command,
        including the pathname.
        Raises NoEmptySpaceForLoadCommandError() if there's not enough space in the Mach-O header to add a new command.
        Note: This will invalidate the binary's code signature, if present.
        TODO(PT): Deprecate and move to MachoBinaryWriter
        """
        if not self.is_64bit:
            raise RuntimeError("Inserting load commands is only support on 64-bit binaries")
        if self.is_swap:
            raise RuntimeError("Unsupported endianness")

        load_cmd = DylibCommand()
        load_cmd.cmd = MachoLoadCommands.LC_LOAD_DYLIB
        sizeof_dylib_struct = sizeof(DylibStruct())
        sizeof_load_cmd = sizeof(load_cmd)
        # XXX(PT): The size of the command should be 0x20 + max(len(dylib_path), 0x20). Found experimentally.
        dylib_path_bytes = bytes(dylib_path, "utf8")
        load_cmd.cmdsize = sizeof_load_cmd + max(len(dylib_path_bytes), 0x20)
        # Align the size on an 8-byte boundary
        if load_cmd.cmdsize % 8:
            load_cmd.cmdsize = (load_cmd.cmdsize + 8) & ~(8 - 1)

        load_cmd.dylib = DylibStruct()
        dylib_name = LcStrUnion()
        # The name starts after the size of the LC_LOAD_DYLIB struct, which is 0x18 bytes
        dylib_name.offset = sizeof_dylib_struct
        load_cmd.dylib.name = dylib_name
        load_cmd.dylib.timestamp = 0x0
        load_cmd.dylib.current_version = 0x0
        load_cmd.dylib.compatibility_version = 0x0

        # TODO(PT): It'd be nice to get a modified MachoBinary with a context manager.
        # This way, all the modifications can be performed on the same binary, without creating a copy
        # for each change.
        # TODO(PT): Alternatively, assignments like `self.header.ncmds += 1` should update the backing binary.
        # This would also make binary modifications easier.

        # Check that there is enough emtpy space before the start of __text to insert a load command
        load_commands_end = self.header.sizeof + self.header.sizeofcmds
        string_end = load_commands_end + load_cmd.cmdsize
        text_section = self.section_with_name("__text", "__TEXT")
        if text_section and string_end >= text_section.offset:
            raise NoEmptySpaceForLoadCommandError()

        # Add the load command to the end of the Mach-O header
        modified_binary = self.write_struct(load_cmd, load_commands_end)
        # Write the dylib path directly after the load cmd
        modified_binary = modified_binary.write_bytes(dylib_path_bytes, load_commands_end + sizeof_dylib_struct)

        # Increase mh_header->ncmds by 1
        bumped_ncmds = self.header.ncmds + 1
        # This field is a 32-bit little-endian encoded int
        bumped_ncmds_bytes = bumped_ncmds.to_bytes(4, "little")
        # The ncmds field is located at the binary head, plus the offset of the field into the mh_header structure
        ncmds_address = MachoHeaderStruct._64_BIT_STRUCT.ncmds.offset
        modified_binary = modified_binary.write_bytes(bumped_ncmds_bytes, ncmds_address)

        # Increase mh_header->sizeofcmds by the size of the new load command
        bumped_sizeofcmds = self.header.sizeofcmds + load_cmd.cmdsize
        # This field is a 32-bit little-endian encoded int
        bumped_sizeofcmds_bytes = bumped_sizeofcmds.to_bytes(4, "little")
        # The sizeofcmds field is located at the binary head, plus the offset of the field into the mh_header structure
        sizeofcmds_address = MachoHeaderStruct._64_BIT_STRUCT.sizeofcmds.offset
        modified_binary = modified_binary.write_bytes(bumped_sizeofcmds_bytes, sizeofcmds_address)

        # All done
        return modified_binary

    def write_binary(self, path: Path) -> None:
        """Write the in-memory Mach-O slice to the provided path.
        TODO(PT): Deprecate and move to MachoBinaryWriter
        """
        # Pass 'x' so the call will throw an exception if the path already exists
        with open(path, "xb") as out_file:
            out_file.write(self._cached_binary)

    @staticmethod
    def write_fat(slices: List["MachoBinary"], path: Path) -> None:
        """Write a list of Mach-O slices into a FAT file at the provided path.
        TODO(PT): Deprecate and move to MachoBinaryWriter
        """
        from strongarm.macho.macho_definitions import MachArch, MachoFatArch, MachoFatHeader

        if any(x.is_swap for x in slices):
            raise RuntimeError("Unsupported endianness")

        # Write the FAT header
        fat_header = MachoFatHeader()
        fat_header.magic = MachArch.FAT_MAGIC.value
        fat_header.nfat_arch = len(slices)
        file_data = bytearray(bytes(fat_header))

        # Write a fat-arch structure for each binary slice
        arch_to_binaries: List[Tuple[MachoFatArch, "MachoBinary"]] = []
        page_size = 0x4000
        for binary in slices:
            arch = MachoFatArch()
            arch.cputype = binary.header.cputype
            arch.cpusubtype = binary.header.cpusubtype
            arch.size = len(binary._cached_binary)
            # Experimentally, in a FAT with an armv7 and arm64 slice, the align for both arch's was log2(0x4000)
            arch.align = int(math.log2(page_size))
            arch_to_binaries.append((arch, binary))

        # Figure out where to place each binary in the file
        speculative_data_end = len(file_data) + (sizeof(MachoFatArch) * len(arch_to_binaries))
        for arch, binary in arch_to_binaries:
            # Find the nearest page boundary
            speculative_data_end = (speculative_data_end + page_size) & ~(page_size - 1)
            # Assign this file offset to the slice
            arch.offset = speculative_data_end
            # Add the size of the binary to the data-size marker
            speculative_data_end += arch.size
            # Write this arch entry to the file data
            file_data += bytearray(bytes(arch))

        # Change the endianess of the FAT header from big to little
        # The FAT header is a list of 32-bit ints
        for idx in range(len(file_data))[::4]:
            # Reverse the bytes of this int
            file_data[idx : idx + 4] = bytearray(reversed(file_data[idx : idx + 4]))

        # We now know the final file size, and where each slice should be placed
        # Zero-fill the rest of the file, and copy in each slice
        # Page-align the file-end
        speculative_data_end = (speculative_data_end + page_size) & ~(page_size - 1)
        file_data += bytearray(speculative_data_end - len(file_data))
        for arch, binary in arch_to_binaries:
            file_data[arch.offset : arch.offset + arch.size] = binary._cached_binary

        # The output file has been constructed. Write it to disk
        # Pass 'x' so the call will throw an exception if the path already exists
        with open(path, "xb") as out_file:
            out_file.write(file_data)

    def get_functions(self) -> Set[VirtualMemoryPointer]:
        """Get a list of the function entry points defined in LC_FUNCTION_STARTS. This includes objective-c methods.

        Returns: A list of VirtualMemoryPointers corresponding to each function's entry point.
        """
        # TODO(PT): move read_uleb somewhere else
        from .dyld_info_parser import DyldInfoParser

        if self._functions_list:
            return self._functions_list

        # Cannot do anything without LC_FUNCTIONS_START
        if not self._function_starts_cmd:
            return set()

        functions_list = set()

        fs_start = self._function_starts_cmd.dataoff
        fs_size = self._function_starts_cmd.datasize
        fs_uleb = self.get_contents_from_address(fs_start, fs_size)

        address = int(self.get_virtual_base())

        idx = 0
        while idx < fs_size:
            address_delta, idx = DyldInfoParser.read_uleb(fs_uleb, idx)

            address += address_delta
            func_entry = VirtualMemoryPointer(address)
            functions_list.add(func_entry)

        self._functions_list = functions_list
        return self._functions_list

    def dylib_id(self) -> Optional[str]:
        """If the binary contains an LC_ID_DYLIB load command, return the pathname which the binary represents.
        """
        if not self._id_dylib_cmd:
            return None

        dylib_name_addr = (
            self._id_dylib_cmd.binary_offset + self._id_dylib_cmd.dylib.name.offset + self.get_virtual_base()
        )
        dylib_name = self.get_full_string_from_start_address(dylib_name_addr)
        if not dylib_name:
            dylib_name = "<unknown dylib>"
        return dylib_name

    def get_minimum_deployment_target(self) -> Optional[LooseVersion]:
        if not self.__minimum_deployment_target:
            if self._build_version_cmd:
                # X.Y.Z is encoded in nibbles xxxx.yy.zz
                encoded_min_target = self._build_version_cmd.minos
                patch = (encoded_min_target >> (8 * 0)) & 0xFF
                minor = (encoded_min_target >> (8 * 1)) & 0xFF
                major = (encoded_min_target >> (8 * 2)) & 0xFFFF
                self.__minimum_deployment_target = LooseVersion(f"{major}.{minor}.{patch}")
        return self.__minimum_deployment_target

    def get_build_version_platform(self) -> Optional[MachoBuildVersionPlatform]:
        if not self._build_version_cmd:
            return None
        return MachoBuildVersionPlatform(self._build_version_cmd.platform)

    def get_build_tool_versions(self) -> Optional[List[MachoBuildToolVersionStruct]]:
        return self._build_tool_versions



from ctypes import BigEndianStructure, c_uint8, c_uint32
from enum import IntEnum

from strongarm.macho.arch_independent_structs import ArchIndependentStructure


class CodesignBlobTypeEnum(IntEnum):
    """Magic numbers for codesigning blobs
    """

    CSMAGIC_REQUIREMENT = 0xFADE0C00  # single requirement blob
    CSMAGIC_REQUIREMENT_SET = 0xFADE0C01  # requirements vector (internal requirements)
    CSMAGIC_CODE_DIRECTORY = 0xFADE0C02  # CodeDirectory blob
    CSMAGIC_EMBEDDED_SIGNATURE = 0xFADE0CC0  # embedded signature data
    CSMAGIC_DETACHED_SIGNATURE = 0xFADE0CC1  # multi-arch collection of embedded signatures
    CSMAGIC_EMBEDDED_ENTITLEMENTS = 0xFADE7171  # embedded entitlements
    CSMAGIC_BLOBWRAPPER = 0xFADE0B01  # CMS signature, "among other things" from the source code


class CSBlobStruct(BigEndianStructure):
    """Basic CodeSign blob structure. These fields shared by all CodeSign blob structures.
    """

    _fields_ = [("magic", c_uint32), ("length", c_uint32)]


class CSSuperblobStruct(BigEndianStructure):
    _fields_ = [*CSBlobStruct._fields_, ("index_entry_count", c_uint32)]


class CSCodeDirectoryStruct(BigEndianStructure):
    _fields_ = [
        *CSBlobStruct._fields_,
        ("version", c_uint32),
        ("flags", c_uint32),
        ("hash_offset", c_uint32),
        ("identifier_offset", c_uint32),
        ("special_slots_count", c_uint32),
        ("code_slots_count", c_uint32),
        ("code_limit", c_uint32),
        ("hash_size", c_uint8),
        ("hash_type", c_uint8),
        ("platform", c_uint8),
        ("page_size", c_uint8),
        ("unused", c_uint32),
        ("scatter_offset", c_uint32),
        ("team_offset", c_uint32),
    ]


class CSBlobIndexStruct(BigEndianStructure):
    _fields_ = [("type", c_uint32), ("offset", c_uint32)]


class CSBlob(ArchIndependentStructure):
    _32_BIT_STRUCT = CSBlobStruct
    _64_BIT_STRUCT = CSBlobStruct


class CSSuperblob(ArchIndependentStructure):
    _32_BIT_STRUCT = CSSuperblobStruct
    _64_BIT_STRUCT = CSSuperblobStruct


class CSCodeDirectory(ArchIndependentStructure):
    _32_BIT_STRUCT = CSCodeDirectoryStruct
    _64_BIT_STRUCT = CSCodeDirectoryStruct


class CSBlobIndex(ArchIndependentStructure):
    _32_BIT_STRUCT = CSBlobIndexStruct
    _64_BIT_STRUCT = CSBlobIndexStruct



import functools
import logging
import pathlib
import shutil
import sqlite3
import tempfile
import time
from contextlib import closing
from ctypes import sizeof
from dataclasses import dataclass
from typing import TYPE_CHECKING, Any, Callable, Dict, Iterable, List, Optional, Set, Tuple, Type, TypeVar, cast

from capstone import CS_ARCH_ARM64, CS_MODE_ARM, Cs, CsInsn
from more_itertools import pairwise

from strongarm.macho.arch_independent_structs import CFString32, CFString64, CFStringStruct
from strongarm.macho.dyld_info_parser import DyldBoundSymbol
from strongarm.macho.macho_binary import InvalidAddressError, MachoBinary
from strongarm.macho.macho_definitions import VirtualMemoryPointer
from strongarm.macho.macho_imp_stubs import MachoImpStubsParser
from strongarm.macho.macho_string_table_helper import MachoStringTableHelper
from strongarm.macho.objc_runtime_data_parser import (
    ObjcCategory,
    ObjcClass,
    ObjcProtocol,
    ObjcRuntimeDataParser,
    ObjcSelector,
)

if TYPE_CHECKING:
    from strongarm.objc import ObjcFunctionAnalyzer, ObjcMethodInfo


_T = TypeVar("_T")


ANALYZER_SQL_SCHEMA = """
    CREATE TABLE function_boundaries(
        entry_point INT NOT NULL UNIQUE,
        end_address INT NOT NULL UNIQUE,
        CHECK (entry_point < end_address)
    );
    CREATE TABLE basic_blocks(
        entry_point INT NOT NULL,
        start_address INT NOT NULL,
        end_address INT NOT NULL,
        CHECK(entry_point <= start_address),
        CHECK(start_address < end_address),
        UNIQUE(entry_point, start_address, end_address)
    );

    CREATE TABLE function_calls(
        destination_address INT,
        caller_address INT,
        caller_func_start_address INT
    );

    CREATE TABLE objc_msgSends(
        destination_address INT,
        caller_address INT,
        caller_func_start_address INT,
        class_name TEXT,
        selector TEXT
    );

    CREATE TABLE named_callable_symbols(
        is_imported INT,
        address INT,
        symbol_name TEXT
    );

    CREATE TABLE string_xrefs(
        string_literal TEXT,
        accessor_address INT,
        accessor_func_start_address INT
    );
"""


class DisassemblyFailedError(Exception):
    """Raised when Capstone fails to disassemble a bytecode sequence.
    """


@dataclass(order=True, frozen=True)
class CallerXRef:
    destination_addr: VirtualMemoryPointer
    caller_addr: VirtualMemoryPointer
    caller_func_start_address: VirtualMemoryPointer


@dataclass(order=True, frozen=True)
class ObjcMsgSendXref(CallerXRef):
    class_name: Optional[str]
    selector: Optional[str]


@dataclass
class CallableSymbol:
    """A locally-defined function or externally-defined imported function
    """

    address: VirtualMemoryPointer
    is_imported: bool
    symbol_name: str


CallableT = TypeVar("CallableT", bound=Callable)


def _requires_xrefs_computed(func: CallableT) -> CallableT:
    @functools.wraps(func)
    def wrap(self: "MachoAnalyzer", *args: Any, **kwargs: Any) -> Any:
        if not self._has_computed_xrefs:
            logging.info(
                f"called {func.__name__} before XRefs were computed for {self.binary.path.name}, computing now..."
            )
            self._build_xref_database()
        return func(self, *args, **kwargs)

    return cast(CallableT, wrap)


class cached_property(object):
    """A property whose value is computed only once.
    Used as a < py3.8 alternative to @functools.cached_property
    Avoiding @functools.lru_cache as they would keep-alive the MachoAnalyzer forever. See:
    https://bugs.python.org/issue19859
    Implementation copied from:
    https://github.com/pallets/werkzeug/blob/0e1b8c4fe598725b343085c5a9a867e90b966db6/werkzeug/utils.py#L35-L73
    """

    def __init__(self, func: Callable) -> None:
        self.__name__ = func.__name__
        self.__module__ = func.__module__
        self.__doc__ = func.__doc__
        self.func = func

    def __get__(self, obj: Any, _type: Type = None) -> Any:
        if obj is None:
            return self
        value = obj.__dict__.get(self.__name__, None)
        if value is None:
            value = self.func(obj)
            obj.__dict__[self.__name__] = value
        return value


class MachoAnalyzer:
    # This class does expensive one-time cross-referencing operations
    # Therefore, we want only one instance to exist for any MachoBinary
    # Thus, the preferred interface for getting an instance of this class is MachoAnalyzer.get_analyzer(binary),
    # which utilizes this cache
    # XXX(PT): These references live to process termination, or until clear_cache() is called
    _ANALYZER_CACHE: Dict[MachoBinary, "MachoAnalyzer"] = {}

    def __init__(self, binary: MachoBinary) -> None:
        self.binary = binary
        self.cs = Cs(CS_ARCH_ARM64, CS_MODE_ARM)
        self.cs.detail = True

        # Each __stubs function calls a single dyld stub address, which has a corresponding DyldBoundSymbol.
        # Map of each __stub function to the associated name of the DyldBoundSymbol
        self._imported_symbol_addresses_to_names: Dict[VirtualMemoryPointer, str] = {}

        self.crossref_helper = MachoStringTableHelper(binary)
        self.imported_symbols = self.crossref_helper.imported_symbols

        self.imp_stubs = MachoImpStubsParser(binary, self.cs).imp_stubs
        self._objc_helper: Optional[ObjcRuntimeDataParser] = None
        self._objc_method_list: List[ObjcMethodInfo] = []

        # Use a temporary database to store cross-referenced data. This provides constant-time lookups for things like
        # finding all the calls to a particular function.
        self._has_computed_xrefs = False
        self._db_tempdir = pathlib.Path(tempfile.mkdtemp())
        self._db_path = self._db_tempdir / "strongarm.db"
        self._db_handle = sqlite3.connect(self._db_path.as_posix())
        cursor = self._db_handle.executescript(ANALYZER_SQL_SCHEMA)
        with self._db_handle:
            cursor.close()

        self._build_callable_symbol_index()
        self._build_function_boundaries_index()

        self._cfstring_to_stringref_map = self._build_cfstring_map()
        self._cstring_to_stringref_map = self._build_cstring_map()

        # Done setting up, store this analyzer in class cache
        MachoAnalyzer._ANALYZER_CACHE[binary] = self

    def __repr__(self) -> str:
        return f"<MachoAnalyzer binary={self.binary.path.as_posix()}>"

    @_requires_xrefs_computed
    def calls_to(self, address: VirtualMemoryPointer) -> List[CallerXRef]:
        """Return the list of code-locations within the binary which branch to the provided address.
        """
        xrefs_cursor = self._db_handle.execute(
            "SELECT * from function_calls WHERE destination_address=?", (int(address),)
        )
        return [CallerXRef(x[0], x[1], x[2]) for x in xrefs_cursor]

    @_requires_xrefs_computed
    def objc_calls_to(
        self, objc_class_names: List[str], objc_selectors: List[str], requires_class_and_sel_found: bool,
    ) -> List[ObjcMsgSendXref]:
        """Return the list of code-locations in the binary which invoke _objc_msgSend with any of the provided
        classes or selectors.
        This also covers iOS 13.5+'s optimized calls that bypass _objc_msgSend, such as _objc_alloc_init.

        If requires_class_and_sel_found is set, a call-site will only be yielded if one of the
        classes and one of the selectors are messaged in the same call.
        Otherwise, a call-site will be yielded if one of the classes *or* one of the selectors are messaged
        at a call site.
        """
        classes_int_list = ", ".join(f'"{x}"' for x in objc_class_names)
        selectors_int_list = ", ".join(f'"{x}"' for x in objc_selectors)

        # Do we require the class and selector being messaged to both be messaged at the same call site?
        query_predicate = "AND" if requires_class_and_sel_found else "OR"
        query = (
            f"SELECT * from objc_msgSends"
            f" WHERE class_name IN ({classes_int_list}) {query_predicate} selector IN ({selectors_int_list})"
        )
        objc_calls_cursor = self._db_handle.execute(query)
        return [ObjcMsgSendXref(x[0], x[1], x[2], x[3], x[4]) for x in objc_calls_cursor]

    def _compute_function_basic_blocks(
        self, entry_point: VirtualMemoryPointer, end_address: VirtualMemoryPointer
    ) -> Iterable[Tuple[int, int]]:
        # PT: This implicitly links against the capstone shared library,
        # and if capstone is not installed correctly it will raise an ImportError.
        # Report this in a clearer way so the user can see exactly what went wrong.
        try:
            from strongarm_dataflow.dataflow import compute_function_basic_blocks_fast
        except ImportError as e:
            if "libcapstone" in str(e):
                import sys

                print("\ncapstone 4.x could not be found, is the capstone backend installed?\n")
                sys.exit(1)
            raise

        bytecode = self.binary.get_content_from_virtual_address(
            virtual_address=entry_point, size=end_address - entry_point
        )
        basic_block_starts = compute_function_basic_blocks_fast(bytecode, entry_point)
        # Convert basic-block starts to [start, end] pairs
        return pairwise(x for x in basic_block_starts)

    def get_basic_block_boundaries(
        self, entry_point: VirtualMemoryPointer
    ) -> List[Tuple[VirtualMemoryPointer, VirtualMemoryPointer]]:
        """Given the function starting at the provided address, return the list of (start_addr, end_addr) basic blocks.
        """
        cursor = self._db_handle.execute(
            "SELECT start_address, end_address FROM basic_blocks WHERE entry_point=?", (entry_point,)
        )
        with closing(cursor):
            return [(VirtualMemoryPointer(x[0]), VirtualMemoryPointer(x[1])) for x in cursor]

    def _build_function_boundaries_index(self) -> None:
        """Iterate all the entry points listed in the binary metadata and compute the end-of-function address for each.
        The end-of-function address for each entry point is then stored in a DB table.

        To compute function boundaries, each function's basic blocks are determined. The end-address is then the
        final address in the final basic block.
        """
        cursor = self._db_handle.cursor()
        sorted_entry_points = sorted(self.get_functions())

        # Computing a function boundaries uses the next entry point address as a hint. For the last entry point in the
        # binary, use the end of the section as the hint.
        try:
            last_entry = sorted_entry_points[-1]
        except IndexError:
            pass
        else:
            section = self.binary.section_for_address(last_entry)
            assert section is not None and section.end_address >= last_entry
            sorted_entry_points.append(VirtualMemoryPointer(section.end_address))

        for entry_point, end_address in pairwise(sorted_entry_points):
            # The end address of the function is the last instruction in the last basic block
            basic_blocks = [x for x in self._compute_function_basic_blocks(entry_point, end_address)]
            # If we found a function with no code, just skip it
            # This can happen in the assembly unit tests, where we insert a jump to a dummy __text label
            if len(basic_blocks) == 0:
                continue
            end_address = VirtualMemoryPointer(max((bb_end for _, bb_end in basic_blocks)))
            cursor.execute(
                "INSERT INTO function_boundaries (entry_point, end_address) VALUES (?, ?)", (entry_point, end_address)
            )
            cursor.executemany(
                "INSERT INTO basic_blocks VALUES (?, ?, ?)", [(entry_point, t[0], t[1]) for t in basic_blocks]
            )

        with self._db_handle:
            cursor.close()

    @cached_property
    def _objc_msgSend_addr(self) -> Optional[VirtualMemoryPointer]:
        objc_msgsend_symbol = self.callable_symbol_for_symbol_name("_objc_msgSend")
        if not objc_msgsend_symbol:
            return None
        return VirtualMemoryPointer(objc_msgsend_symbol.address)

    @cached_property
    def _objc_fastpath_ptrs_to_selector_names(self) -> Dict[VirtualMemoryPointer, str]:
        # Some special selectors have a fast-path, _objc_opt_<selector>, that was added in iOS 13
        # The _objc_opt_* call is emitted instead of _objc_msgSend when the NSObject implementation will be called.
        # Found with: $ nm "/usr/lib/libobjc.A.dylib" | grep "objc_opt"
        # There are also functions in this family that don't begin with "_opt", such as _objc_alloc.
        fastpath_function_names_to_selector_names = {
            "_objc_opt_class": "class",
            "_objc_opt_isKindOfClass": "isKindOfClass:",
            "_objc_opt_new": "new",
            "_objc_opt_respondsToSelector": "respondsToSelector:",
            "_objc_opt_self": "self",
            "_objc_alloc": "alloc",
            "_objc_alloc_init": "init",
        }
        fastpath_funcptrs_to_selector_names: Dict[VirtualMemoryPointer, str] = {}
        for func_name, selector_name in fastpath_function_names_to_selector_names.items():
            # A callable symbol is only present if the function has been used in this binary
            sym = self.callable_symbol_for_symbol_name(func_name)
            if not sym:
                continue
            fastpath_funcptrs_to_selector_names[sym.address] = selector_name
        return fastpath_funcptrs_to_selector_names

    def _build_xref_database(self) -> None:
        """Iterate all the code in the binary and populate the following DB tables:
        * function_calls
        * objc_msgSends
        * string_xrefs
        """
        from strongarm_dataflow.dataflow import build_xref_database_fast

        if self._has_computed_xrefs:
            logging.error("Already computed xrefs, why was _build_xref_database called again?")
            return

        start_time = time.time()
        logging.debug(f"{self.binary.path} computing call XRefs...")

        objc_function_family = list(self._objc_fastpath_ptrs_to_selector_names.keys())
        if self._objc_msgSend_addr:
            objc_function_family.append(self._objc_msgSend_addr)

        build_xref_database_fast(
            self,
            self.binary.path.as_posix(),
            self._db_path.as_posix(),
            self.binary.get_virtual_base(),
            self.binary.get_file_offset(),
            self._objc_msgSend_addr,
            objc_function_family,
            list(self.get_function_boundaries()),
        )

        self._has_computed_xrefs = True
        end_time = time.time()
        logging.debug(f"Finding xrefs took {end_time - start_time} seconds")

    @classmethod
    def clear_cache(cls) -> None:
        """Delete cached MachoAnalyzer's
        This can be used when you are finished analyzing a binary set and don't want to retain the cached data in memory
        """
        for binary, analyzer in cls._ANALYZER_CACHE.items():
            logging.debug(f"Deleting db {analyzer._db_path}...")
            analyzer._db_handle.close()
            shutil.rmtree(analyzer._db_tempdir.as_posix())

        cls._ANALYZER_CACHE.clear()

    @property
    def objc_helper(self) -> ObjcRuntimeDataParser:
        if not self._objc_helper:
            self._objc_helper = ObjcRuntimeDataParser(self.binary)
        return self._objc_helper

    @classmethod
    def get_analyzer(cls, binary: MachoBinary) -> "MachoAnalyzer":
        """Get a cached analyzer for a given MachoBinary
        """
        if binary in cls._ANALYZER_CACHE:
            # There exists a MachoAnalyzer for this binary - use it instead of making a new one
            return cls._ANALYZER_CACHE[binary]
        return MachoAnalyzer(binary)

    def method_info_for_entry_point(self, entry_point: VirtualMemoryPointer) -> Optional["ObjcMethodInfo"]:
        # TODO(PT): This should return any symbol name, not just Obj-C methods
        from strongarm.objc.objc_analyzer import ObjcMethodInfo

        for objc_cls in self.objc_classes():
            for sel in objc_cls.selectors:
                if sel.implementation == entry_point:
                    return ObjcMethodInfo(objc_cls, sel, sel.implementation)
        return None

    def objc_classes(self) -> List[ObjcClass]:
        """Return the List of classes and categories implemented within the binary
        """
        return self.objc_helper.classes

    def objc_categories(self) -> List[ObjcCategory]:
        """Return the List of categories implemented within the app
        """
        all_classes = self.objc_classes()
        categories: List[ObjcCategory] = [c for c in all_classes if isinstance(c, ObjcCategory)]
        return categories

    def get_conformed_protocols(self) -> List[ObjcProtocol]:
        """Return the List of protocols to which code within the binary conforms
        """
        return self.objc_helper.protocols

    @property
    def dyld_bound_symbols(self) -> Dict[VirtualMemoryPointer, DyldBoundSymbol]:
        """Return a Dict of each imported dyld stub to the corresponding symbol to be bound at runtime.
        """
        return self.binary.dyld_bound_symbols

    @property
    def imp_stubs_to_symbol_names(self) -> Dict[VirtualMemoryPointer, str]:
        """Return a Dict of callable implementation stubs to the names of the imported symbols they correspond to.
        """
        if self._imported_symbol_addresses_to_names:
            return self._imported_symbol_addresses_to_names

        symbol_name_map = {}
        stubs = self.imp_stubs
        stub_map = self.dyld_bound_symbols

        unnamed_stub_count = 0
        for stub in stubs:
            if stub.destination in stub_map:
                symbol_name = stub_map[stub.destination].name
                symbol_name_map[stub.address] = symbol_name
            else:
                # add in stub which is not backed by a named symbol
                # a stub contained in the __stubs section that was not backed by a named symbol was first
                # encountered in com.intuit.mobilebanking01132.app/PlugIns/CMA Balance Widget.appex/CMA Balance Widget
                name = f"unnamed_stub_{unnamed_stub_count}"
                unnamed_stub_count += 1
                symbol_name_map[stub.destination] = name

        self._imported_symbol_addresses_to_names = symbol_name_map
        return symbol_name_map

    @cached_property
    def imported_symbols_to_symbol_names(self) -> Dict[VirtualMemoryPointer, str]:
        """Return a Dict of imported symbol pointers to their names.
        These symbols are not necessarily callable, but may rather be imported classes, for example.
        Inverse of MachoAnalyzer.imported_symbol_names_to_pointers()
        """
        return {addr: x.name for addr, x in self.dyld_bound_symbols.items()}

    @cached_property
    def imported_symbol_names_to_pointers(self) -> Dict[str, VirtualMemoryPointer]:
        """Return a Dict of imported symbol names to their pointers.
        These symbols are not necessarily callable.
        Inverse of MachoAnalyzer.imported_symbol_names_to_pointers()

        NOTE: This API can lose information! A bound symbol of the same name may be bound to multiple locations,
        and this API only retains one of those locations. For example, ___CFConstantStringClassReference is bound to
        the first field of every CFStringStruct in the binary. This API probably shouldn't be used for this reason...
        """
        return {x.name: addr for addr, x in self.dyld_bound_symbols.items()}

    @property
    def exported_symbol_pointers_to_names(self) -> Dict[VirtualMemoryPointer, str]:
        """Return a Dict of pointers to exported symbol definitions to their symbol names.
        Inverse of MachoAnalyzer.exported_symbol_names_to_pointers()
        """
        return self.crossref_helper.exported_symbols

    @cached_property
    def exported_symbol_names_to_pointers(self) -> Dict[str, VirtualMemoryPointer]:
        """Return a Dict of exported symbol names to pointers to their definitions.
        Inverse of MachoAnalyzer.exported_symbols_to_symbol_names()
        """
        return {y: x for x, y in self.exported_symbol_pointers_to_names.items()}

    def exported_symbol_name_for_address(self, address: VirtualMemoryPointer) -> Optional[str]:
        """Return the symbol name for the provided address, or None if the address is not a named exported symbol.
        """
        if address in self.exported_symbol_pointers_to_names:
            return self.exported_symbol_pointers_to_names[address]
        return None

    def symbol_name_for_branch_destination(self, branch_address: VirtualMemoryPointer) -> str:
        """Get the associated symbol name for a given branch destination
        """
        if branch_address in self.imp_stubs_to_symbol_names:
            return self.imp_stubs_to_symbol_names[branch_address]
        raise RuntimeError(f"Unknown branch destination {hex(branch_address)}. Is this a local branch?")

    def disassemble_region(self, start_address: VirtualMemoryPointer, size: int) -> List[CsInsn]:
        """Disassemble the executable code in a given region into a list of CsInsn objects
        """
        func_str = bytes(self.binary.get_content_from_virtual_address(virtual_address=start_address, size=size))
        instructions = [instr for instr in self.cs.disasm(func_str, start_address)]
        if not len(instructions):
            raise DisassemblyFailedError(f"Failed to disassemble code at {hex(start_address)}:{hex(size)}")
        return instructions

    def get_function_instructions(self, start_address: VirtualMemoryPointer) -> List[CsInsn]:
        """Get a list of disassembled instructions for the function beginning at start_address
        """
        end_address = self.get_function_end_address(start_address)

        if end_address is None:
            raise RuntimeError(f"No function with start address {start_address} found.")

        instructions = self.disassemble_region(start_address, end_address - start_address)
        return instructions

    def imp_for_selref(self, selref_ptr: VirtualMemoryPointer) -> Optional[VirtualMemoryPointer]:
        selector = self.objc_helper.selector_for_selref(selref_ptr)
        if not selector:
            return None
        return selector.implementation

    def selector_for_selref(self, selref_ptr: VirtualMemoryPointer) -> Optional[ObjcSelector]:
        return self.objc_helper.selector_for_selref(selref_ptr)

    def selector_for_selector_literal(self, selref_ptr: VirtualMemoryPointer) -> Optional[ObjcSelector]:
        return self.objc_helper.selector_for_selector_literal(selref_ptr)

    def get_method_imp_addresses(self, selector: str) -> List[VirtualMemoryPointer]:
        """Given a selector, return a list of virtual addresses corresponding to the start of each IMP for that SEL
        """
        return self.objc_helper.get_method_imp_addresses(selector)

    def get_imps_for_sel(self, selector: str) -> List["ObjcFunctionAnalyzer"]:
        """Retrieve a list of the disassembled function data for every implementation of a provided selector
        Args:
            selector: The selector name who's implementations should be found

        Returns:
            A list of ObjcFunctionAnalyzers corresponding to each found implementation of the provided selector.
        """
        from strongarm.objc import ObjcFunctionAnalyzer  # noqa: F811

        implementation_analyzers = []
        imp_addresses = self.get_method_imp_addresses(selector)
        for imp_start in imp_addresses:
            imp_instructions = self.get_function_instructions(imp_start)
            function_analyzer = ObjcFunctionAnalyzer(self.binary, imp_instructions)
            implementation_analyzers.append(function_analyzer)
        return implementation_analyzers

    def get_objc_methods(self) -> List["ObjcMethodInfo"]:
        """Get a List of ObjcMethodInfo's representing all ObjC methods implemented in the Mach-O.
        """
        from strongarm.objc import ObjcMethodInfo  # type: ignore

        if self._objc_method_list:
            return self._objc_method_list
        method_list = []
        for objc_class in self.objc_classes():
            for objc_sel in objc_class.selectors:
                imp_addr = objc_sel.implementation

                info = ObjcMethodInfo(objc_class, objc_sel, imp_addr)
                method_list.append(info)
        self._objc_method_list = method_list
        return self._objc_method_list

    def get_functions(self) -> Set[VirtualMemoryPointer]:
        """Get a list of the function entry points defined in LC_FUNCTION_STARTS. This includes objective-c methods.

        Returns: A list of VirtualMemoryPointers corresponding to each function's entry point.
        """
        return self.binary.get_functions()

    def get_function_boundaries(self) -> Set[Tuple[VirtualMemoryPointer, VirtualMemoryPointer]]:
        cursor = self._db_handle.execute("SELECT entry_point, end_address FROM function_boundaries")

        with closing(cursor):
            return {(VirtualMemoryPointer(a), VirtualMemoryPointer(b)) for a, b in cursor}

    def get_function_end_address(self, entry_point: VirtualMemoryPointer) -> Optional[VirtualMemoryPointer]:
        cursor = self._db_handle.execute(
            "SELECT end_address FROM function_boundaries WHERE entry_point = ?", (entry_point,)
        )

        with closing(cursor):
            results = cursor.fetchone()

        if results is None:
            return None

        return VirtualMemoryPointer(results[0])

    @cached_property
    def class_for_class_pointer_map(self) -> Dict[VirtualMemoryPointer, ObjcClass]:
        return {VirtualMemoryPointer(x.raw_struct.binary_offset): x for x in self.objc_classes()}

    def class_name_for_class_pointer(self, classref: VirtualMemoryPointer) -> Optional[str]:
        """Given a classref, return the name of the class.
        This method will handle classes implemented within the binary and imported classes.
        """
        # Did the caller provide a classref for an imported class?
        local_class = self.imported_symbols_to_symbol_names.get(classref)
        if local_class:
            return local_class

        # The class is implemented within the binary and has an associated ObjcClass object
        # We could have been passed either a classref pointer in __objc_classrefs, or the direct address of
        # an __objc_data structure in __objc_const. Try both variants to search for the associated class.

        # First, check if we were provided with the address of an __objc_data struct in __objc_data representing
        # the class.
        local_class = self.class_for_class_pointer_map.get(classref)
        if local_class:
            return local_class.name

        # Then, check if we were passed a classref pointer in __objc_classrefs
        try:
            dereferenced_classref = VirtualMemoryPointer(self.binary.read_word(classref))
        except InvalidAddressError:
            # Invalid classref
            return None

        local_class = self.class_for_class_pointer_map.get(dereferenced_classref)
        if local_class:
            return local_class.name

        # Invalid classref
        return None

    def classref_for_class_name(self, class_name: str) -> Optional[VirtualMemoryPointer]:
        """Given a class name, try to find a classref for it.
        """
        classrefs = [
            addr
            for addr, name in self.imported_symbols_to_symbol_names.items()
            if name == class_name and self.binary.section_name_for_address(addr) == "__objc_classrefs"
        ]
        if len(classrefs):
            return classrefs[0]

        # TODO(PT): this is expensive! We should do one analysis step of __objc_classrefs and create a map.
        classref_locations, classref_destinations = self.binary.read_pointer_section("__objc_classrefs")

        # is it a local class?
        class_locations = [x.raw_struct.binary_offset for x in self.objc_classes() if x.name == class_name]
        if not len(class_locations):
            # unknown class name
            return None
        class_location = VirtualMemoryPointer(class_locations[0])

        if class_location not in classref_destinations:
            # unknown class name
            return None

        classref_index = classref_destinations.index(class_location)
        return classref_locations[classref_index]

    def selref_for_selector_name(self, selector_name: str) -> Optional[VirtualMemoryPointer]:
        return self.objc_helper.selref_for_selector_name(selector_name)

    @_requires_xrefs_computed
    def strings(self) -> Set[str]:
        """Return a list containing every string in the binary
        """
        # Gather strings from various sections
        all_strings = set()
        for section_name in ["__cstring", "__objc_methname", "__objc_methtype", "__objc_classname", "__const"]:
            section_strings = self._strings_in_section(section_name)
            all_strings.update(section_strings)

        # Gather strings found via Xrefs
        c = self._db_handle.cursor()
        xref_strings_rows = c.execute("SELECT string_literal from string_xrefs").fetchall()
        xref_strings = [xref_strings_row[0] for xref_strings_row in xref_strings_rows]
        all_strings.update(set(xref_strings))

        return all_strings

    def get_cstrings(self) -> Set[str]:
        """Return the list of strings in the binary's __cstring section.
        """
        # TODO(PT): This is SLOW and WASTEFUL!!!
        # These transformations should be done ONCE on initial analysis!
        # This method should cache its result.
        return self._strings_in_section("__cstring")

    def _build_cstring_map(self) -> Dict[str, VirtualMemoryPointer]:
        strings_section = self.binary.section_with_name("__cstring", "__TEXT")
        if not strings_section:
            return {}

        strings_base = strings_section.address
        strings_content = self.binary.get_bytes(strings_section.offset, strings_section.size)

        string_to_stringrefs = {}
        # split into characters (string table is packed and each entry is terminated by a null character)
        string_table = list(strings_content)
        transformed_strings = MachoStringTableHelper.transform_string_section(string_table)
        for idx, entry in transformed_strings.items():
            # Address is the base of __cstring plus the index of the entry
            stringref_address = VirtualMemoryPointer(strings_base + idx)
            string_to_stringrefs[entry.full_string] = stringref_address
        return string_to_stringrefs

    def _stringref_for_cstring(self, string: str) -> Optional[VirtualMemoryPointer]:
        """Try to find the stringref in __cstrings for a provided C string.
        If the string is not present in the __cstrings section, this method returns None.
        """
        if string not in self._cstring_to_stringref_map:
            return None
        return self._cstring_to_stringref_map[string]

    def _build_cfstring_map(self) -> Dict[str, VirtualMemoryPointer]:
        cfstrings_section = self.binary.section_with_name("__cfstring", "__DATA")
        if not cfstrings_section:
            cfstrings_section = self.binary.section_with_name("__cfstring", "__DATA_CONST")
            if not cfstrings_section:
                return {}

        sizeof_cfstring = sizeof(CFString64) if self.binary.is_64bit else sizeof(CFString32)
        cfstrings_base = cfstrings_section.address

        cfstring_to_stringrefs = {}
        cfstrings_count = int((cfstrings_section.end_address - cfstrings_section.address) / sizeof_cfstring)
        for i in range(cfstrings_count):
            cfstring_addr = cfstrings_base + (i * sizeof_cfstring)
            cfstring = self.binary.read_struct_with_rebased_pointers(cfstring_addr, CFStringStruct, virtual=True)
            literal = self.binary.read_string_at_address(cfstring.literal)
            if literal:
                cfstring_to_stringrefs[literal] = VirtualMemoryPointer(cfstring_addr)
        return cfstring_to_stringrefs

    def _stringref_for_cfstring(self, string: str) -> Optional[VirtualMemoryPointer]:
        """Try to find the stringref in __cfstrings for a provided Objective-C string literal.
        If the string is not present in the __cfstrings section, this method returns None.
        """
        if string not in self._cfstring_to_stringref_map:
            return None
        return self._cfstring_to_stringref_map[string]

    def stringref_for_string(self, string: str) -> Optional[VirtualMemoryPointer]:
        """Try to find the stringref for a provided string.
        If the string is not present in the binary, this method returns None.

        If you are looking for a C string, pass the string with no additional formatting.
        If you are looking for an Objective-C string literal (CFStringRef), enclose your string in @"".
        """
        is_cfstring = False
        if string.startswith('@"'):
            if not string.endswith('"'):
                raise RuntimeError(f"incorrectly formatted ObjC string literal {string}")

            is_cfstring = True
            # trim the @" prefix and the " suffix
            string = string[2:-1]

        if is_cfstring:
            return self._stringref_for_cfstring(string)
        return self._stringref_for_cstring(string)

    @functools.lru_cache(64)
    def callable_symbol_for_address(self, branch_destination: VirtualMemoryPointer) -> Optional[CallableSymbol]:
        """Retrieve information about a callable branch destination.
        It's the caller's responsibility to provide a valid branch destination with a symbol associated with it.
        """
        c = self._db_handle.cursor()
        symbols = c.execute("SELECT * from named_callable_symbols WHERE address=?", (branch_destination,)).fetchall()
        if not len(symbols):
            return None
        assert len(symbols) == 1, f"Found more than 1 symbol at {branch_destination}?"
        symbol_data = symbols[0]

        return CallableSymbol(
            is_imported=bool(symbol_data[0]), address=VirtualMemoryPointer(symbol_data[1]), symbol_name=symbol_data[2]
        )

    def callable_symbol_for_symbol_name(self, symbol_name: str) -> Optional[CallableSymbol]:
        """Retrieve information about a name within the imported or exported symbols tables.
        It's the caller's responsibility to provide a valid callable symbol name.
        """
        c = self._db_handle.cursor()
        symbols = c.execute("SELECT * from named_callable_symbols WHERE symbol_name=?", (symbol_name,)).fetchall()
        if not len(symbols):
            return None
        assert len(symbols) == 1, f"Found more than 1 symbol named {symbol_name}?"
        symbol_data = symbols[0]

        return CallableSymbol(
            is_imported=bool(symbol_data[0]), address=VirtualMemoryPointer(symbol_data[1]), symbol_name=symbol_data[2]
        )

    @_requires_xrefs_computed
    def string_xrefs_to(self, string_literal: str) -> List[Tuple[VirtualMemoryPointer, VirtualMemoryPointer]]:
        """Retrieve each code location that loads the provided (C or CF) string.
        Returns a tuple of (function entry point, instruction which completes the string load)
        """
        c = self._db_handle.cursor()
        xrefs_query = c.execute(
            "SELECT accessor_func_start_address, accessor_address from string_xrefs WHERE string_literal=?",
            (string_literal,),
        ).fetchall()
        string_xrefs = [(VirtualMemoryPointer(x[0]), VirtualMemoryPointer(x[1])) for x in xrefs_query]
        return string_xrefs

    @_requires_xrefs_computed
    def strings_in_func(self, func_addr: VirtualMemoryPointer) -> List[Tuple[VirtualMemoryPointer, str]]:
        """Fetch the list of strings referenced by the provided function.
        Returns a tuple of (instruction that completes the string load, loaded string literal)
        """
        c = self._db_handle.cursor()
        xrefs: Iterable[Tuple[int, str]] = c.execute(
            "SELECT accessor_address, string_literal from string_xrefs WHERE accessor_func_start_address=?",
            (func_addr,),
        )
        string_loads = [(VirtualMemoryPointer(x[0]), x[1]) for x in xrefs]
        return string_loads

    def _build_callable_symbol_index(self) -> None:
        """Build a database index for every callable symbol to symbol name.
        This index includes both imported and exported symbols.
        """
        c = self._db_handle.cursor()

        # Process __imp_stubs
        imp_stub_addr_and_symbol_name = (
            (stub_addr, sym_name) for stub_addr, sym_name in self.imp_stubs_to_symbol_names.items()
        )
        c.executemany("INSERT INTO named_callable_symbols VALUES (1, ?, ?)", imp_stub_addr_and_symbol_name)

        # Process the symbols defined in the binary
        callable_addr_and_sym_name = (
            (callable_addr, sym_name) for callable_addr, sym_name in self.exported_symbol_pointers_to_names.items()
        )
        c.executemany("INSERT INTO named_callable_symbols VALUES (0, ?, ?)", callable_addr_and_sym_name)

        self._db_handle.commit()

    def _strings_in_section(self, section_name: str) -> Set[str]:
        """Fetch the list of strings located inside the provided section
        """
        discovered_strings = set()
        string_section = self.binary.section_with_name(section_name, "__TEXT")
        if string_section:
            strings_content = self.binary.get_bytes(string_section.offset, string_section.size)
            transformed_strings = MachoStringTableHelper.transform_string_section(list(strings_content))
            discovered_strings = set((x.full_string for x in transformed_strings.values()))
        return discovered_strings



import logging
import os
import subprocess

from setuptools import find_packages, setup
from setuptools.command.build_ext import build_ext
from setuptools.command.develop import develop
from setuptools.command.install import install

from strongarm import __url__, __version__


def install_capstone():
    platform = getattr(os.uname(), "sysname", None)
    logging.info(f"Installing Capstone for platform: {platform}")

    if platform == "Darwin":
        logging.info("Installing Capstone backend from brew...")
        subprocess.run(["brew", "install", "capstone"])
    elif platform == "Linux":
        logging.info("Installing Capstone backend from apt-get...")
        subprocess.run(["apt-get", "update"])
        subprocess.run(
            [
                "apt-get",
                "install",
                "libcapstone4",
                "libcapstone-dev",
                "sqlite3",
                "libsqlite3-dev",
                "-y",
                "--allow-unauthenticated",
            ]
        )
    else:
        # Let's not make this a fatal error, as the user may be able to install Capstone on their own
        logging.warning(f"Unknown platform: {platform}")
        logging.warning("You must install the capstone backend before using strongarm")


# https://stackoverflow.com/questions/19569557/pip-not-picking-up-a-custom-install-cmdclass
class InstallCapstoneBuildExtCmd(build_ext):
    def run(self) -> None:
        install_capstone()
        super().run()


class InstallCapstoneInstallCmd(install):
    def run(self) -> None:
        install_capstone()
        super().run()


class InstallCapstoneDevelopCmd(develop):
    def run(self) -> None:
        install_capstone()
        super().run()


# Ensure our logs when installing Capstone show up
logging.basicConfig(level=logging.INFO)

setup(
    name="strongarm-ios",
    version=__version__,
    description="Mach-O/ARM64 analyzer",
    author="Data Theorem",
    url=__url__,
    packages=find_packages(exclude=["tests"]),
    install_requires=["capstone", "more_itertools", "strongarm_dataflow==2.1.5"],
    package_data={"strongarm": ["py.typed"]},
    data_files=[("", ["LICENSE.txt"])],
    cmdclass={
        "build_ext": InstallCapstoneBuildExtCmd,
        "install": InstallCapstoneInstallCmd,
        "develop": InstallCapstoneDevelopCmd,
    },
)



import logging
from ctypes import c_uint32, sizeof
from pathlib import Path
from typing import Dict, List, Optional, Tuple, Type, TypeVar

from _ctypes import Structure

from strongarm.macho.macho_binary import MachoBinary
from strongarm.macho.macho_definitions import (
    DyldSharedCacheHeader,
    DyldSharedCacheImageInfo,
    DyldSharedFileMapping,
    MachArch,
    StaticFilePointer,
    VirtualMemoryPointer,
    VMProtFlags,
)

_StructureT = TypeVar("_StructureT", bound=Structure)


class DyldSharedCacheParser:
    """Top-level mechanism for parsing a dyld_shared_cache

    Useful links:
        https://opensource.apple.com/source/dyld/dyld-195.6/launch-cache/dsc_iterator.cpp.auto.html
        https://opensource.apple.com/source/dyld/dyld-655.1.1/launch-cache/dyld_cache_format.h.auto.html
    """

    # TODO(PT): Eventually, we could have a generic file-loader which shares some logic of MachoParser/DSCParser

    _DSC_MAGIC = [MachArch.DYLD_SHARED_CACHE_MAGIC]

    def __init__(self, path: Path) -> None:
        self.path = path

        # DSC's are split into 3 "mappings", or segments:
        # Mapping 0 is the executable segment. __TEXT of embedded binaries is placed here
        # Mapping 1 is the writable segment. __DATA/writable data of embedded binaries is placed here
        # Mapping 2 is the readonly segment. __LINKEDIT data (such as symbol tables) is placed here
        # This attribute stores the parsed mapping structures
        self.segment_mappings: List[DyldSharedFileMapping] = []

        # DSC's store a number of system dylibs.
        # This attribute stores the path of an embedded dylib to the virtual mapping of its __TEXT segment
        # In other words, the value for each path is a tuple of:
        # - The VM pointer to the image's Mach-O header
        # - The VM pointer to the end-address of the Mach-O's __TEXT segment
        self.embedded_binary_info: Dict[Path, Tuple[VirtualMemoryPointer, VirtualMemoryPointer]] = {}

        self._parse()

    @property
    def file_magic(self) -> int:
        """Read file magic
        """
        return c_uint32.from_buffer(bytearray(self.get_bytes(StaticFilePointer(0), sizeof(c_uint32)))).value

    def get_bytes(self, offset: StaticFilePointer, size: int) -> bytes:
        """Read a region of bytes from the input file
        Args:
            offset: Offset within file to begin reading from
            size: Maximum number of bytes to read
        Returns:
            Byte list representing contents of file at provided address
        """
        with open(str(self.path), "rb") as binary_file:
            binary_file.seek(offset)
            return binary_file.read(size)

    def read_struct(self, file_offset: StaticFilePointer, struct_type: Type[_StructureT]) -> _StructureT:
        """Given a file offset, return the structure it describes
        Args:
            file_offset: Address from where to read the bytes
            struct_type: Structure subclass
        Returns:
            struct_type loaded from the pointed address
        """
        data = bytearray(self.get_bytes(file_offset, sizeof(struct_type)))
        return struct_type.from_buffer(data)  # type: ignore

    def _read_static_c_string(self, start_address: StaticFilePointer) -> Optional[str]:
        """Return a string containing the bytes from start_address up to the next NULL character
        This method will return None if the specified address does not point to a UTF-8 encoded string
        """
        max_len = 16
        symbol_name_characters = []
        found_null_terminator = False

        while not found_null_terminator:
            name_bytes = self.get_bytes(start_address, max_len)
            # search for null terminator in this content
            for ch in name_bytes:
                if ch == 0x00:
                    found_null_terminator = True
                    break
                symbol_name_characters.append(ch)

            # do we need to keep searching for the end of the symbol name?
            if not found_null_terminator:
                # since we read [start_address:start_address + max_len], trim that from search space
                start_address += max_len
                # double search space for next iteration
                max_len *= 2
            else:
                # read full string!
                try:
                    symbol_name = bytearray(symbol_name_characters).decode("UTF-8")
                    return symbol_name
                except UnicodeDecodeError:
                    # if decoding the string failed, we may have been passed an address which does not actually
                    # point to a string
                    return None
        return None

    def _parse(self) -> None:
        # Read the shared-cache header
        self.header = self.read_struct(StaticFilePointer(0), DyldSharedCacheHeader)

        logging.debug(f"Cache magic: {self.header.magic.decode()}")
        logging.debug(f"First mapping: {hex(self.header.mappingOffset)}")
        logging.debug(f"Mapping count: {self.header.mappingCount}")
        logging.debug(f"First image: {hex(self.header.imagesOffset)}")
        logging.debug(f"Image count: {self.header.imagesCount}")
        logging.debug(f"Memory base: {hex(self.header.dyldBaseAddress)}")
        logging.debug(f"Codesign base: {hex(self.header.codeSignOffset)}")

        self._parse_dsc_mappings()
        self._parse_embedded_binaries()

    def _parse_dsc_mappings(self) -> None:
        """Populates self.segment_translations based on the mappings reported by the DSC header.
        """
        # We expect exactly: an executable mapping, a writable mapping, a readonly mapping
        # Verify this expectation
        assert self.header.mappingCount == 3
        expected_vm_protections = [
            VMProtFlags.VM_PROT_READ | VMProtFlags.VM_PROT_EXECUTE,
            VMProtFlags.VM_PROT_READ | VMProtFlags.VM_PROT_WRITE,
            VMProtFlags.VM_PROT_READ,
        ]

        # Parse the DSC mappings
        mapping_off = self.header.mappingOffset
        for mapping_idx in range(self.header.mappingCount):
            mapping_struct = self.read_struct(StaticFilePointer(mapping_off), DyldSharedFileMapping)
            mapping_off += sizeof(DyldSharedCacheImageInfo)

            virt_addr = VirtualMemoryPointer(mapping_struct.address)
            virt_end = virt_addr + mapping_struct.size

            static_addr = StaticFilePointer(mapping_struct.file_offset)
            prot = mapping_struct.max_prot

            logging.debug(f"Mapping [{mapping_idx}]: [{virt_addr} - {virt_end}] @ {static_addr}, prot = {prot}")

            # Verify the permissions of this mapping are as we expect
            assert prot == expected_vm_protections[mapping_idx], f"{hex(prot)} {expected_vm_protections[mapping_idx]}"

            self.segment_mappings.append(mapping_struct)

    def _parse_embedded_binaries(self) -> None:
        """Populates self.embedded_binary_info based on the images reported by the DSC header.
        """
        # Parse the embedded binaries within the DSC
        image_off = self.header.imagesOffset
        for image_idx in range(self.header.imagesCount):
            image_struct = self.read_struct(StaticFilePointer(image_off), DyldSharedCacheImageInfo)
            image_off += sizeof(DyldSharedCacheImageInfo)

            # Example: /System/Library/Frameworks/CoreFoundation.framework/CoreFoundation
            embedded_binary_path_str = self._read_static_c_string(image_struct.pathFileOffset)
            if not embedded_binary_path_str:
                file_offset = image_off - sizeof(DyldSharedCacheImageInfo)
                raise ValueError(f"Failed to read an image name for image struct @ {hex(file_offset)}")
            embedded_binary_path = Path(embedded_binary_path_str)

            vm_addr = VirtualMemoryPointer(image_struct.address)
            # To calculate the size of this image, we need to look at the next image's address
            # Except for the last image, which doesn't have an image after it
            if image_idx == self.header.imagesCount - 1:
                mapping_end = VirtualMemoryPointer(self.segment_mappings[0].address + self.segment_mappings[0].size)
                image_size = mapping_end - vm_addr
            else:
                next_image = self.read_struct(StaticFilePointer(image_off), DyldSharedFileMapping)
                image_size = next_image.address - image_struct.address

            vm_end = vm_addr + image_size
            self.embedded_binary_info[Path(embedded_binary_path)] = (vm_addr, vm_end)

    def translate_virtual_address_to_static(self, vm_addr: VirtualMemoryPointer) -> StaticFilePointer:
        """Given a pointer within the DSC's virtual address mappings, return the file pointer to the same data.
        """
        # Find the mapping which contains the provided address
        for mapping in self.segment_mappings:
            if mapping.address <= vm_addr < mapping.address + mapping.size:
                offset_into_segment = vm_addr - mapping.address
                return StaticFilePointer(mapping.file_offset + offset_into_segment)
        raise ValueError(f"Could not find address within DSC address space: {vm_addr}")

    def get_embedded_binary(self, binary_path: Path) -> "DyldSharedCacheBinary":
        """Given a path to a binary embedded in the DSC, retrieve & parse the embedded binary
        """
        if binary_path not in self.embedded_binary_info:
            raise ValueError(f"DSC does not contain {binary_path}")

        text_vm_start, text_vm_end = self.embedded_binary_info[binary_path]
        text_size = text_vm_end - text_vm_start
        logging.debug(f"Parsing DSC image {binary_path} @ [{text_vm_start}, {text_vm_end}]")

        static_addr = self.translate_virtual_address_to_static(text_vm_start)
        image_bytes = self.get_bytes(static_addr, text_size)

        return DyldSharedCacheBinary(self, binary_path, static_addr, image_bytes)

    def image_for_text_address(self, address: VirtualMemoryPointer) -> Path:
        """Given a virtual memory address of __TEXT content, return the embedded image which contains it
        """
        for path, text_region in self.embedded_binary_info.items():
            text_vm_start, text_vm_end = text_region
            if text_vm_start <= address < text_vm_end:
                return path
        raise ValueError(f"No embedded __TEXT segment contains {address}")


class DyldSharedCacheBinary(MachoBinary):
    """A special Mach-O binary which exists within a dyld_shared_cache.
    DSC binaries are different from Mach-O binaries in that its segments are dispersed across the DSC file.
    Its __TEXT is in DSC.segment_mappings[0]. This means that all the header-structures parsed by the MachoBinary
    intializer are within the __TEXT buffer.
    However, these structures may point to things in DSC.segment_mappings[1] or DSC.segment_mappings[2].
    For example, a symtab command of a DSC binary will contain pointers to __LINKEDIT data within the DSC's __LINKEDIT.
    Thus, these binaries are supported like so:
    - The DSC binary retains a reference to the global DSC
    - When the DSC binary is initialized, MachoBinary only reads the __TEXT segment bytes
    - When get_bytes() is called, it checks whether the address is within __TEXT.
        If not, it will read the data from the global DSC, with an optional flag to disable the translation.
        The translation must be disabled for a few reads, such as symbol-table parsing, as the DSC has pre-
        translated these values.
    """

    def __init__(
        self, dsc_parser: "DyldSharedCacheParser", path: Path, file_offset: StaticFilePointer, binary_data: bytes
    ) -> None:
        self.dyld_shared_cache_parser = dsc_parser
        self.dyld_shared_cache_file_offset = file_offset
        super().__init__(path, binary_data)

    def file_offset_for_virtual_address(self, virtual_address: VirtualMemoryPointer) -> StaticFilePointer:
        # Translate into the global DSC file
        return self.dyld_shared_cache_parser.translate_virtual_address_to_static(virtual_address)

    def get_bytes(self, offset: StaticFilePointer, size: int, _translate_addr_to_file: bool = True) -> bytearray:
        # There are two possibilities: The requested data is "binary-local", meaning it's within the __TEXT buffer
        # backing this object. Or, the requested data is somewhere within the global DSC.
        # It would be clear which is the case from the calling context. For example, if the pointer comes from
        # a symbol table, it's probably in the DSC-global __LINKEDIT mapping. In contrast, if the pointer represents
        # the bytecode for some function, it's probably in the __TEXT mapping. We don't want to provide this context
        # from every get_bytes caller, so try to determine what data is being requested here.
        # If offset+size refers to an address outside the local image, translate and read from the global DSC.
        # Otherwise, don't translate and read directly from the global DSC.
        if offset + size > self.dyld_shared_cache_file_offset + len(self._cached_binary):
            logging.debug(f"Reading from addr outside __TEXT: {offset}")
            # This address is outside the binary's buffer. If translation was disabled, an assumption has been violated
            assert _translate_addr_to_file, f"Must translate addr outside __TEXT: {offset}"

        else:
            if _translate_addr_to_file:
                offset += self.dyld_shared_cache_file_offset
            else:
                logging.debug(f"Translation explicitly disabled, direct read of {offset}")

        return bytearray(self.dyld_shared_cache_parser.get_bytes(offset, size))



from .arch_independent_structs import (
    ArchIndependentStructure,
    CFStringStruct,
    DylibCommandStruct,
    MachoDyldInfoCommandStruct,
    MachoDysymtabCommandStruct,
    MachoEncryptionInfoStruct,
    MachoHeaderStruct,
    MachoLinkeditDataCommandStruct,
    MachoLoadCommandStruct,
    MachoNlistStruct,
    MachoSectionRawStruct,
    MachoSegmentCommandStruct,
    MachoSymtabCommandStruct,
    ObjcCategoryRawStruct,
    ObjcClassRawStruct,
    ObjcDataRawStruct,
    ObjcMethodListStruct,
    ObjcMethodStruct,
    ObjcProtocolListStruct,
    ObjcProtocolRawStruct,
)
from .dyld_info_parser import BindOpcode, DyldBoundSymbol, DyldInfoParser
from .dyld_shared_cache import DyldSharedCacheBinary, DyldSharedCacheParser
from .macho_analyzer import CallerXRef, MachoAnalyzer, ObjcMsgSendXref
from .macho_binary import (
    BinaryEncryptedError,
    InvalidAddressError,
    LoadCommandMissingError,
    MachoBinary,
    MachoSection,
    MachoSegment,
    NoEmptySpaceForLoadCommandError,
)
from .macho_definitions import (
    CPU_TYPE,
    HEADER_FLAGS,
    NLIST_NTYPE,
    NTYPE_VALUES,
    CFString32,
    CFString64,
    DyldSharedCacheHeader,
    DyldSharedCacheImageInfo,
    DyldSharedFileMapping,
    DylibCommand,
    DylibStruct,
    LcStrUnion,
    MachArch,
    MachoDyldInfoCommand,
    MachoDysymtabCommand,
    MachoEncryptionInfo32Command,
    MachoEncryptionInfo64Command,
    MachoFatArch,
    MachoFatHeader,
    MachoFileType,
    MachoHeader32,
    MachoHeader64,
    MachoLinkeditDataCommand,
    MachoLoadCommand,
    MachoNlist32,
    MachoNlist64,
    MachoNlistUn,
    MachoSection32Raw,
    MachoSection64Raw,
    MachoSegmentCommand32,
    MachoSegmentCommand64,
    MachoSymtabCommand,
    ObjcCategoryRaw32,
    ObjcCategoryRaw64,
    ObjcClassRaw32,
    ObjcClassRaw64,
    ObjcDataRaw32,
    ObjcDataRaw64,
    ObjcMethod32,
    ObjcMethod64,
    ObjcMethodList,
    ObjcProtocolList32,
    ObjcProtocolList64,
    ObjcProtocolRaw32,
    ObjcProtocolRaw64,
    StaticFilePointer,
    VirtualMemoryPointer,
    swap32,
)
from .macho_imp_stubs import MachoImpStub, MachoImpStubsParser
from .macho_load_commands import MachoLoadCommands
from .macho_parse import ArchitectureNotSupportedError, MachoParser
from .macho_string_table_helper import MachoStringTableEntry, MachoStringTableHelper
from .objc_runtime_data_parser import (
    ObjcCategory,
    ObjcClass,
    ObjcProtocol,
    ObjcRuntimeDataParser,
    ObjcSelector,
    ObjcSelref,
)

__all__ = [
    "ArchIndependentStructure",
    "CFStringStruct",
    "DylibCommandStruct",
    "MachoDyldInfoCommandStruct",
    "MachoDysymtabCommandStruct",
    "MachoEncryptionInfoStruct",
    "MachoHeaderStruct",
    "MachoLinkeditDataCommandStruct",
    "MachoLoadCommandStruct",
    "MachoNlistStruct",
    "MachoSectionRawStruct",
    "MachoSegmentCommandStruct",
    "MachoSymtabCommandStruct",
    "ObjcCategoryRawStruct",
    "ObjcClassRawStruct",
    "ObjcDataRawStruct",
    "ObjcMethodListStruct",
    "ObjcMethodStruct",
    "ObjcProtocolListStruct",
    "ObjcProtocolRawStruct",
    "BindOpcode",
    "DyldBoundSymbol",
    "DyldInfoParser",
    "DyldSharedCacheBinary",
    "DyldSharedCacheParser",
    "CallerXRef",
    "MachoAnalyzer",
    "ObjcMsgSendXref",
    "BinaryEncryptedError",
    "InvalidAddressError",
    "LoadCommandMissingError",
    "MachoBinary",
    "MachoSection",
    "MachoSegment",
    "NoEmptySpaceForLoadCommandError",
    "CPU_TYPE",
    "HEADER_FLAGS",
    "NLIST_NTYPE",
    "NTYPE_VALUES",
    "CFString32",
    "CFString64",
    "DyldSharedCacheHeader",
    "DyldSharedCacheImageInfo",
    "DyldSharedFileMapping",
    "DylibCommand",
    "DylibStruct",
    "LcStrUnion",
    "MachArch",
    "MachoDyldInfoCommand",
    "MachoDysymtabCommand",
    "MachoEncryptionInfo32Command",
    "MachoEncryptionInfo64Command",
    "MachoFatArch",
    "MachoFatHeader",
    "MachoFileType",
    "MachoHeader32",
    "MachoHeader64",
    "MachoLinkeditDataCommand",
    "MachoLoadCommand",
    "MachoNlist32",
    "MachoNlist64",
    "MachoNlistUn",
    "MachoSection32Raw",
    "MachoSection64Raw",
    "MachoSegmentCommand32",
    "MachoSegmentCommand64",
    "MachoSymtabCommand",
    "ObjcCategoryRaw32",
    "ObjcCategoryRaw64",
    "ObjcClassRaw32",
    "ObjcClassRaw64",
    "ObjcDataRaw32",
    "ObjcDataRaw64",
    "ObjcMethod32",
    "ObjcMethod64",
    "ObjcMethodList",
    "ObjcProtocolList32",
    "ObjcProtocolList64",
    "ObjcProtocolRaw32",
    "ObjcProtocolRaw64",
    "StaticFilePointer",
    "VirtualMemoryPointer",
    "swap32",
    "MachoStringTableEntry",
    "MachoStringTableHelper",
    "ArchitectureNotSupportedError",
    "MachoParser",
    "MachoLoadCommands",
    "MachoImpStub",
    "MachoImpStubsParser",
    "ObjcCategory",
    "ObjcClass",
    "ObjcProtocol",
    "ObjcRuntimeDataParser",
    "ObjcSelector",
    "ObjcSelref",
]



import struct
from ctypes import Structure, Union, c_char, c_char_p, c_int16, c_int32, c_uint8, c_uint16, c_uint32, c_uint64
from enum import IntEnum
from typing import TypeVar

_BasePointerT = TypeVar("_BasePointerT", bound="_BasePointer")


class _BasePointer(int):
    def __add__(self: _BasePointerT, other: int) -> _BasePointerT:
        return type(self)(super().__add__(other))

    def __sub__(self: _BasePointerT, other: int) -> _BasePointerT:
        return self.__class__(super().__sub__(other))

    def __mul__(self: _BasePointerT, other: int) -> _BasePointerT:
        return self.__class__(super().__mul__(other))

    def __truediv__(self: _BasePointerT, other: int) -> _BasePointerT:
        return self.__class__(super().__truediv__(other))

    def __floordiv__(self: _BasePointerT, other: int) -> _BasePointerT:
        return self.__class__(super().__floordiv__(other))

    def __str__(self) -> str:
        return hex(self)

    def __repr__(self) -> str:
        return hex(self)


class StaticFilePointer(_BasePointer):
    """A pointer analogous to a file offset within the Mach-O
    """

    def __str__(self) -> str:
        return f"Phys[{super().__str__()}]"

    def __repr__(self) -> str:
        return f"Phys[{super().__repr__()}]"


class VirtualMemoryPointer(_BasePointer):
    """A pointer representing a virtual memory location within the Mach-O
    """


def swap32(i: int) -> int:
    """Reverse the bytes of a little-endian integer representation ie (3) -> 50331648"""
    return struct.unpack("<I", struct.pack(">I", i))[0]


class MachArch(IntEnum):
    MH_MAGIC = 0xFEEDFACE
    MH_CIGAM = 0xCEFAEDFE
    MH_MAGIC_64 = 0xFEEDFACF
    MH_CIGAM_64 = 0xCFFAEDFE

    FAT_MAGIC = 0xCAFEBABE
    FAT_CIGAM = 0xBEBAFECA

    MH_CPU_ARCH_ABI64 = 0x01000000
    MH_CPU_TYPE_ARM = 12
    MH_CPU_TYPE_ARM64 = MH_CPU_TYPE_ARM | MH_CPU_ARCH_ABI64

    MH_CPU_SUBTYPE_PTRAUTH_ABI = 0x80000000
    MH_CPU_SUBTYPE_ARM64E = 0x02

    DYLD_SHARED_CACHE_MAGIC = 0x646C7964  # b'dyld'


class VMProtFlags(IntEnum):
    # https://opensource.apple.com/source/xnu/xnu-1504.7.4/osfmk/mach/vm_prot.h.auto.html
    VM_PROT_NONE = 0 << 0
    VM_PROT_READ = 1 << 0
    VM_PROT_WRITE = 1 << 1
    VM_PROT_EXECUTE = 1 << 2


class CPU_TYPE(IntEnum):
    ARMV7 = 0
    ARM64 = 1
    UNKNOWN = 2


class MachoFileType(IntEnum):
    MH_OBJECT = 1  # relocatable object file
    MH_EXECUTE = 2  # demand paged executable file
    MH_FVMLIB = 3  # fixed VM shared library file
    MH_CORE = 4  # core file
    MH_PRELOAD = 5  # preloaded executable file
    MH_DYLIB = 6  # dynamically bound shared library
    MH_DYLINKER = 7  # dynamic link editor
    MH_BUNDLE = 8  # dynamically bound bundle file
    MH_DYLIB_STUB = 9  # shared library stub for static linking only, no section contents
    MH_DSYM = 10  # shared library stub for static
    MH_KEXT_BUNDLE = 11  # x86_64 kext


class MachoHeader32(Structure):
    _fields_ = [
        ("magic", c_uint32),
        ("cputype", c_uint32),
        ("cpusubtype", c_uint32),
        ("filetype", c_uint32),
        ("ncmds", c_uint32),
        ("sizeofcmds", c_uint32),
        ("flags", c_uint32),
    ]


class MachoHeader64(Structure):
    _fields_ = [
        ("magic", c_uint32),
        ("cputype", c_uint32),
        ("cpusubtype", c_uint32),
        ("filetype", c_uint32),
        ("ncmds", c_uint32),
        ("sizeofcmds", c_uint32),
        ("flags", c_uint32),
        ("reserved", c_uint32),
    ]


class MachoLoadCommand(Structure):
    _fields_ = [("cmd", c_uint32), ("cmdsize", c_uint32)]


class MachoSegmentCommand32(Structure):
    _fields_ = [
        *MachoLoadCommand._fields_,
        ("segname", c_char * 16),
        ("vmaddr", c_uint32),
        ("vmsize", c_uint32),
        ("fileoff", c_uint32),
        ("filesize", c_uint32),
        ("maxprot", c_uint32),
        ("initprot", c_uint32),
        ("nsects", c_uint32),
        ("flags", c_uint32),
    ]


class MachoSegmentCommand64(Structure):
    _fields_ = [
        *MachoLoadCommand._fields_,
        ("segname", c_char * 16),
        ("vmaddr", c_uint64),
        ("vmsize", c_uint64),
        ("fileoff", c_uint64),
        ("filesize", c_uint64),
        ("maxprot", c_uint32),
        ("initprot", c_uint32),
        ("nsects", c_uint32),
        ("flags", c_uint32),
    ]


class MachoSection32Raw(Structure):
    _fields_ = [
        ("sectname", c_char * 16),
        ("segname", c_char * 16),
        ("addr", c_uint32),
        ("size", c_uint32),
        ("offset", c_uint32),
        ("align", c_uint32),
        ("reloff", c_uint32),
        ("nreloc", c_uint32),
        ("flags", c_uint32),
        ("reserved1", c_uint32),
        ("reserved2", c_uint32),
    ]


class MachoSection64Raw(Structure):
    _fields_ = [
        ("sectname", c_char * 16),
        ("segname", c_char * 16),
        ("addr", c_uint64),
        ("size", c_uint64),
        ("offset", c_uint32),
        ("align", c_uint32),
        ("reloff", c_uint32),
        ("nreloc", c_uint32),
        ("flags", c_uint32),
        ("reserved1", c_uint32),
        ("reserved2", c_uint32),
        ("reserved3", c_uint32),
    ]


class MachoDysymtabCommand(Structure):
    """Python representation of struct dysymtab_command

    Definition found in <mach-o/loader.h>
    """

    _fields_ = [
        *MachoLoadCommand._fields_,
        ("ilocalsym", c_uint32),
        ("nlocalsym", c_uint32),
        ("iextdefsym", c_uint32),
        ("nextdefsym", c_uint32),
        ("iundefsym", c_uint32),
        ("nundefsym", c_uint32),
        ("tocoff", c_uint32),
        ("ntoc", c_uint32),
        ("modtaboff", c_uint32),
        ("nmodtab", c_uint32),
        ("extrefsymoff", c_uint32),
        ("nextrefsyms", c_uint32),
        ("indirectsymoff", c_uint32),
        ("nindirectsyms", c_uint32),
        ("extreloff", c_uint32),
        ("nextrel", c_uint32),
        ("locreloff", c_uint32),
        ("nlocrel", c_uint32),
    ]


class MachoSymtabCommand(Structure):
    """Python representation of struct symtab_command

    Definition found in <mach-o/loader.h>
    """

    _fields_ = [
        *MachoLoadCommand._fields_,
        ("symoff", c_uint32),
        ("nsyms", c_uint32),
        ("stroff", c_uint32),
        ("strsize", c_uint32),
    ]


class MachoDyldInfoCommand(Structure):
    """Python representation of struct dyld_info_command

    Definition found in <mach-o/loader.h>
    """

    _fields_ = [
        *MachoLoadCommand._fields_,
        ("rebase_off", c_uint32),
        ("rebase_size", c_uint32),
        ("bind_off", c_uint32),
        ("bind_size", c_uint32),
        ("weak_bind_off", c_uint32),
        ("weak_bind_size", c_uint32),
        ("lazy_bind_off", c_uint32),
        ("lazy_bind_size", c_uint32),
        ("export_off", c_uint32),
        ("export_size", c_uint32),
        ("weak_bind_size", c_uint32),
        ("weak_bind_size", c_uint32),
    ]


class MachoLinkeditDataCommand(Structure):
    """Python representation of struct linkedit_data_command

    Definition found in <mach-o/loader.h>
    """

    _fields_ = [*MachoLoadCommand._fields_, ("dataoff", c_uint32), ("datasize", c_uint32)]


class MachoBuildVersionCommand(Structure):
    """Python representation of struct build_version_command

    Definition found in <mach-o/loader.h>
    """

    _fields_ = [
        *MachoLoadCommand._fields_,
        ("platform", c_uint32),
        ("minos", c_uint32),
        ("sdk", c_uint32),
        ("ntools", c_uint32),
    ]


class MachoBuildVersionPlatform(IntEnum):
    MACOS = 1
    IOS = 2
    TVOS = 3
    WATCHOS = 4
    BRIDGEOS = 5
    IOSMAC = 6
    MACCATALYST = 6
    IOSSIMULATOR = 7
    TVOSSIMULATOR = 8
    WATCHOSSIMULATOR = 9
    DRIVERKIT = 10


class MachoBuildToolVersion(Structure):
    """Python representation of struct build_tool_version

    Definition found in <mach-o/loader.h>
    """

    _fields_ = [
        ("tool", c_uint32),
        ("version", c_uint32),
    ]


class MachoBuildTool(IntEnum):
    CLANG = 1
    SWIFT = 2
    LD = 3


class MachoNlistUn(Union):
    """Python representation of union n_un

    Definition found in <mach-o/nlist.h>
    """

    __slots__ = ["n_strx"]
    _fields_ = [("n_strx", c_uint32)]


class MachoNlist32(Structure):
    """Python representation of struct nlist

    Definition found in <mach-o/nlist.h>
    """

    __slots__ = ["n_un", "n_type", "n_sect", "n_desc", "n_value"]
    _fields_ = [
        ("n_un", MachoNlistUn),
        ("n_type", c_uint8),
        ("n_sect", c_uint8),
        ("n_desc", c_int16),
        ("n_value", c_uint32),
    ]


class MachoNlist64(Structure):
    """Python representation of struct nlist_64

    Definition found in <mach-o/nlist.h>
    """

    __slots__ = ["n_un", "n_type", "n_sect", "n_desc", "n_value"]
    _fields_ = [
        ("n_un", MachoNlistUn),
        ("n_type", c_uint8),
        ("n_sect", c_uint8),
        ("n_desc", c_uint16),
        ("n_value", c_uint64),
    ]


class MachoEncryptionInfo32Command(Structure):
    """Python representation of a struct encryption_info_command

    Definition found in <mach-o/loader.h>
    """

    _fields_ = [*MachoLoadCommand._fields_, ("cryptoff", c_uint32), ("cryptsize", c_uint32), ("cryptid", c_uint32)]


class MachoEncryptionInfo64Command(Structure):
    """Python representation of a struct encryption_info_command_64

    Definition found in <mach-o/loader.h>
    """

    _fields_ = [
        *MachoLoadCommand._fields_,
        ("cryptoff", c_uint32),
        ("cryptsize", c_uint32),
        ("cryptid", c_uint32),
        ("pad", c_uint32),
    ]


class MachoFatHeader(Structure):
    """Python representation of a struct fat_header

    Definition found in <mach-o/fat.h>
    """

    _fields_ = [("magic", c_uint32), ("nfat_arch", c_uint32)]


class MachoFatArch(Structure):
    """Python representation of a struct fat_arch

    Definition found in <mach-o/fat.h>
    """

    _fields_ = [
        ("cputype", c_uint32),
        ("cpusubtype", c_uint32),
        ("offset", c_uint32),
        ("size", c_uint32),
        ("align", c_uint32),
    ]


class DyldSharedCacheHeader(Structure):
    # https://opensource.apple.com/source/dyld/dyld-655.1.1/launch-cache/dyld_cache_format.h.auto.html
    _fields_ = [
        ("magic", c_char * 16),  # e.g. "dyld_v1   arm64"
        ("mappingOffset", c_uint32),  # file offset to first shared_file_mapping
        ("mappingCount", c_uint32),  # number of shared_file_mapping entries
        ("imagesOffset", c_uint32),  # file offset to first dyld_cache_image_info
        ("imagesCount", c_uint32),  # number of dyld_cache_image_info entries
        ("dyldBaseAddress", c_uint64),  # base address of dyld when cache was built
        ("codeSignOffset", c_uint64),  # file offset of code signature blob
        ("codeSignSize", c_uint64),  # size of code signature blob
        ("slideInfoOffset", c_uint64),  # file offset of kernel slid info
        ("slideInfoSize", c_uint64),  # size of kernel slid info
        ("localSymbolsOffset", c_uint64),  # file offset where local symbols are stored
        ("localSymbolsSize", c_uint64),  # size of local symbols
        ("uuid", c_char * 16),  # unique value for each shared_cache file
        ("cacheType", c_uint64),  # 0 for dev, 1 for prod
        ("branchPoolsOffset", c_uint32),  # file offset to table of uint64_t pool addresses
        ("branchPoolsSize", c_uint32),  # number of uint64_t entries
        ("accelerateInfoAddr", c_uint64),  # (unslid) address of optimization info
        ("accelerateInfoSize", c_uint64),  # size of optimization info
        ("imagesTextOffset", c_uint64),  # file offset to first dyld_cache_image_text_info
        ("imagesTextCount", c_uint64),  # number of dyld_cache_image_text_info entries
    ]


class DyldSharedFileMapping(Structure):
    _fields_ = [
        ("address", c_uint64),
        ("size", c_uint64),
        ("file_offset", c_uint64),
        ("max_prot", c_uint32),
        ("init_prot", c_uint32),
    ]


class DyldSharedCacheImageInfo(Structure):
    _fields_ = [
        ("address", c_uint64),
        ("modTime", c_uint64),
        ("inode", c_uint64),
        ("pathFileOffset", c_uint32),
        ("pad", c_uint32),
    ]


class NLIST_NTYPE(IntEnum):
    N_STAB = 0xE0  # symbollic debugging entry
    N_PEXT = 0x10  # private external symbol bit
    N_TYPE = 0x0E  # mask for type bits
    N_EXT = 0x01  # external symbol bit


class NTYPE_VALUES(IntEnum):
    N_UNDF = 0x0  # undefined, n_sect == NO_SECT
    N_ABS = 0x2  # absolute, n_sect == NO_SECT
    N_SECT = 0xE  # defined in section n_sect
    N_PBUD = 0xC  # prebound undefined (defined in a dylib)
    N_INDR = 0xA  # indirect


class HEADER_FLAGS(IntEnum):
    NOUNDEFS = 0x1
    INCRLINK = 0x2
    DYLDLINK = 0x4
    BINDATLOAD = 0x8
    PREBOUND = 0x10
    SPLIT_SEGS = 0x20
    LAZY_INIT = 0x40
    TWOLEVEL = 0x80
    FORCE_FLAT = 0x100
    NOMULTIDEFS = 0x200
    NOFIXPREBINDING = 0x400
    PREBINDABLE = 0x800
    ALLMODSBOUND = 0x1000
    SUBSECTIONS_VIA_SYMBOLS = 0x2000
    CANONICAL = 0x4000
    WEAK_DEFINES = 0x8000
    BINDS_TO_WEAK = 0x10000
    ALLOW_STACK_EXECUTION = 0x20000
    ROOT_SAFE = 0x40000
    SETUID_SAFE = 0x80000
    NO_REEXPORTED_DYLIBS = 0x100000
    PIE = 0x200000
    DEAD_STRIPPABLE_DYLIB = 0x400000
    HAS_TLV_DESCRIPTORS = 0x800000
    NO_HEAP_EXECUTION = 0x1000000
    APP_EXTENSION_SAFE = 0x2000000


# Some of these can be found at
# https://opensource.apple.com/source/objc4/objc4-723/runtime/objc-runtime-new.h.auto.html


class ObjcProtocolRaw32(Structure):
    _fields_ = [
        ("isa", c_uint32),
        ("name", c_uint32),
        ("protocols", c_uint32),
        ("required_instance_methods", c_uint32),
        ("required_class_methods", c_uint32),
        ("optional_instance_methods", c_uint32),
        ("optional_class_methods", c_uint32),
        ("instance_properties", c_uint32),
        ("instance_properties", c_uint32),
        ("size", c_uint32),
        ("flags", c_uint32),
    ]


class ObjcProtocolRaw64(Structure):
    _fields_ = [
        ("isa", c_uint64),
        ("name", c_uint64),
        ("protocols", c_uint64),
        ("required_instance_methods", c_uint64),
        ("required_class_methods", c_uint64),
        ("optional_instance_methods", c_uint64),
        ("optional_class_methods", c_uint64),
        ("instance_properties", c_uint64),
        ("instance_properties", c_uint64),
        ("size", c_uint32),
        ("flags", c_uint32),
    ]


class ObjcCategoryRaw32(Structure):
    _fields_ = [
        ("name", c_uint32),
        ("base_class", c_uint32),
        ("instance_methods", c_uint32),
        ("class_methods", c_uint32),
        ("base_protocols", c_uint32),
        ("instance_properties", c_uint32),
    ]


class ObjcCategoryRaw64(Structure):
    _fields_ = [
        ("name", c_uint64),
        ("base_class", c_uint64),
        ("instance_methods", c_uint64),
        ("class_methods", c_uint64),
        ("base_protocols", c_uint64),
        ("instance_properties", c_uint64),
    ]


class ObjcClassRaw32(Structure):
    _fields_ = [
        ("metaclass", c_uint32),
        ("superclass", c_uint32),
        ("cache", c_uint32),
        ("vtable", c_uint32),
        ("data", c_uint32),
    ]


class ObjcClassRaw64(Structure):
    _fields_ = [
        ("metaclass", c_uint64),
        ("superclass", c_uint64),
        ("cache", c_uint64),
        ("vtable", c_uint64),
        ("data", c_uint64),
    ]


class ObjcDataRaw32(Structure):
    _fields_ = [
        ("flags", c_uint32),
        ("instance_start", c_uint32),
        ("instance_size", c_uint32),
        ("ivar_layout", c_uint32),
        ("name", c_uint32),
        ("base_methods", c_uint32),
        ("base_protocols", c_uint32),
        ("ivars", c_uint32),
        ("weak_ivar_layout", c_uint32),
        ("base_properties", c_uint32),
    ]


class ObjcDataRaw64(Structure):
    _fields_ = [
        ("flags", c_uint32),
        ("instance_start", c_uint32),
        ("instance_size", c_uint32),
        ("reserved", c_uint32),
        ("ivar_layout", c_uint64),
        ("name", c_uint64),
        ("base_methods", c_uint64),
        ("base_protocols", c_uint64),
        ("ivars", c_uint64),
        ("weak_ivar_layout", c_uint64),
        ("base_properties", c_uint64),
    ]


class ObjcMethodList(Structure):
    _fields_ = [("flags", c_uint32), ("methcount", c_uint32)]


class ObjcIvarList(Structure):
    _fields_ = [("entsize", c_uint32), ("count", c_uint32)]


class ObjcProtocolList32(Structure):
    _fields_ = [("count", c_uint32)]


class ObjcProtocolList64(Structure):
    _fields_ = [("count", c_uint64)]


class ObjcMethod32(Structure):
    _fields_ = [("name", c_uint32), ("signature", c_uint32), ("implementation", c_uint32)]


class ObjcMethod64(Structure):
    _fields_ = [("name", c_uint64), ("signature", c_uint64), ("implementation", c_uint64)]


class ObjcMethodRelativeData(Structure):
    # Keep the field names the same so that this can be interacted with in the same way as any other ObjcMethodStruct
    # In reality, these fields are closer to: selref_off, signature_off, implementation_off
    # Note that the `name` field instead points to a selref that must be dereferenced to retrieve the name.
    _fields_ = [("name", c_int32), ("signature", c_int32), ("implementation", c_int32)]


class ObjcIvar32(Structure):
    _fields_ = [
        ("offset_ptr", c_uint32),
        ("name", c_uint32),
        ("type", c_uint32),
        ("unknown", c_uint32),
        ("size", c_uint32),
    ]


class ObjcIvar64(Structure):
    _fields_ = [
        ("offset_ptr", c_uint64),
        ("name", c_uint64),
        ("type", c_uint64),
        ("unknown", c_uint32),
        ("size", c_uint32),
    ]


class LcStrUnion(Union):
    _fields_ = [("offset", c_uint32), ("ptr", c_char_p)]


class DylibStruct(Structure):
    _fields_ = [
        ("name", LcStrUnion),
        ("timestamp", c_uint32),
        ("current_version", c_uint32),
        ("compatibility_version", c_uint32),
    ]


class DylibCommand(Structure):
    _fields_ = [*MachoLoadCommand._fields_, ("dylib", DylibStruct)]


class CFString32(Structure):
    _fields_ = [("base", c_uint32), ("flags", c_uint32), ("literal", c_uint32), ("length", c_uint32)]


class CFString64(Structure):
    _fields_ = [("base", c_uint64), ("flags", c_uint64), ("literal", c_uint64), ("length", c_uint64)]


class MachoDyldChainedFixupsHeaderRaw(Structure):
    _fields_ = [
        # 0
        ("fixups_version", c_uint32),
        # Offset of dyld_chained_starts_in_image in chain_data
        ("starts_offset", c_uint32),
        # Offset of imports table in chain_data
        ("imports_offset", c_uint32),
        # Offset of symbol strings in chain_data
        ("symbols_offset", c_uint32),
        # Number of imported symbol names
        ("imports_count", c_uint32),
        # DYLD_CHAINED_IMPORT*
        ("imports_format", c_uint32),
        # 0 => uncompressed, 1 => zlib compressed
        ("symbols_format", c_uint32),
    ]


class MachoDyldChainedImportFormat(IntEnum):
    DYLD_CHAINED_IMPORT = 1
    DYLD_CHAINED_IMPORT_ADDEND = 2
    DYLD_CHAINED_IMPORT_ADDEND64 = 3


class MachoDyldChainedImportRaw(Structure):
    _fields_ = [
        ("lib_ordinal", c_uint32, 8),
        ("weak_import", c_uint32, 1),
        ("name_offset", c_uint32, 23),
    ]


class MachoDyldChainedStartsInImageRaw(Structure):
    _fields_ = [
        ("seg_count", c_uint32),
        # Each entry is offset into this struct for that segment
        # followed by pool of dyld_chain_starts_in_segment data
        # XXX(PT): Although this is declared as uint32_t[1] array in the dyld source, it's actually
        # an array of `seg_count` entries.
        # To avoid hacks that would let ctypes parse this correctly, we manually read the offsets
        # ("seg_info_offset", c_uint32 * 1),
    ]


class MachoDyldChainedStartsInSegmentRaw(Structure):
    # https://docs.python.org/3/library/ctypes.html#structure-union-alignment-and-byte-order
    # XXX(PT): Force alignment to uint16_t, as by default this structure is reported as being 2 bytes too big
    # The correct size is important when we're parsing data that is placed directly after this structure
    _pack_ = 2

    _fields_ = [
        # Size of this (amount kernel needs to copy)
        ("size", c_uint32),
        # 0x1000 or 0x4000
        ("page_size", c_uint16),
        # DYLD_CHAINED_PTR_*
        ("pointer_format", c_uint16),
        # Offset in memory to start of segment
        ("segment_offset", c_uint64),
        # For 32-bit OS, any value beyond this is not a pointer
        ("max_valid_pointer", c_uint32),
        # How many pages are in array
        ("page_count", c_uint16),
        # Each entry is offset in each page of first element in chain
        # or DYLD_CHAINED_PTR_START_NONE if no fixups on page
        # XXX(PT): Variable-length array, see comment on MachoDyldChainedStartsInImageRaw
        # ("page_start", c_uint16),
        # Some 32-bit formats may require multiple starts per page.
        # For those, if high bit is set in page_starts[], then it
        # is index into chain_starts[] which is a list of starts
        # the last of which has the high bit set
        # XXX(PT): Variable-length array, see comment on MachoDyldChainedStartsInImageRaw
        # ("chain_starts", c_uint16),
    ]


class MachoDyldChainedPointerStartType(IntEnum):
    # Used in page_start[] to denote a page with no fixups
    DYLD_CHAINED_PTR_START_NONE = 0xFFFF
    # Used in page_start[] to denote a page which has multiple starts
    DYLD_CHAINED_PTR_START_MULTI = 0x8000
    # Used in chain_starts[] to denote last start in list for page
    DYLD_CHAINED_PTR_START_LAST = 0x8000


class MachoDyldChainedPtr64RebaseRaw(Structure):
    # Used with DYLD_CHAINED_PTR_64/DYLD_CHAINED_PTR_64_OFFSET
    _fields_ = [
        # 64GB max image size (DYLD_CHAINED_PTR_64 => vmAddr, DYLD_CHAINED_PTR_64_OFFSET => runtimeOffset)
        ("target", c_uint64, 36),
        # Top 8 bits set to this (DYLD_CHAINED_PTR_64 => after slide added,
        # DYLD_CHAINED_PTR_64_OFFSET => before slide added)
        ("high8", c_uint64, 8),
        # All zeros
        ("reserved", c_uint64, 7),
        # 4-byte stride
        ("next", c_uint64, 12),
        # == 0
        ("bind", c_uint64, 1),
    ]


class MachoDyldChainedPtr64BindRaw(Structure):
    _fields_ = [
        ("ordinal", c_uint64, 24),
        # 0 thru 255
        ("addend", c_uint64, 8),
        # All zeroes
        ("reserved", c_uint64, 19),
        # 4-byte stride
        ("next", c_uint64, 12),
        # == 1
        ("bind", c_uint64, 1),
    ]
